package cz.aimtec.dci.microservices;


import cz.aimtec.dci.DciApplicationServlet;
import cz.aimtec.dci.DciApplicationWorkspace;
import cz.aimtec.dci.DciCompanyWorkspace;
import cz.aimtec.dci.DciWorkspace;
import cz.aimtec.dci.coreservices.ServiceConsumptionClient;
import cz.aimtec.dci.coreservices.api.model.*;
import cz.aimtec.dci.microservice.discovery.DiscoveryService;
import cz.aimtec.dci.microservices.model.coreservices.ServicesStatusContainer;
import cz.aimtec.dci.services.Service;
import cz.aimtec.dci.services.ServiceEvent;
import cz.aimtec.dci.services.ServiceEventState;
import cz.aimtec.dci.services.SystemServicesDispatcher;
import cz.aimtec.dci.services.technology.tcpip.SubscriptionItem;
import cz.aimtec.util.Convert;
import cz.aimtec.util.datatype.DateTime;
import cz.aimtec.util.exception.ApplicationException;
import java.util.Arrays;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;

/**
 * Core services service client
 *
 * @author	hors
 * @version	21. 11. 2017/hors - created<br>
 */
public class CoreServices extends ServiceConsumptionClient {
	private static final Logger logger = LoggerFactory.getLogger(CoreServices.class);
		
	/**
	 * Basic constructor.
	 *
	 * @param pDiscovery	service discovery provider
	 */
	public CoreServices(DiscoveryService pDiscovery) {
		super(pDiscovery);
	}


	/**
	 * All active services registerred at core services service.
	 *
	 * @param pWorkspace	session workspace
	 * @return				services status container
	 */
	private ServicesStatusContainer getActiveServices(DciWorkspace pWorkspace, String serviceCode) {
		ServicesStatusContainer statusContainer = new ServicesStatusContainer();
		DciCompanyWorkspace companyWorkspace = pWorkspace.getCompanyWorkspace();
		SystemServicesDispatcher companyDispatcher = companyWorkspace.getSystemServicesDispatcher();
		DciApplicationWorkspace applicationWorkspace = DciApplicationWorkspace.getInstance();
		DciApplicationServlet applicationServlet = (DciApplicationServlet) applicationWorkspace.getApplicationServlet();
		SystemServicesDispatcher applicationDispatcher = applicationServlet.getSystemServicesDispatcher();
		
		// ---------------------- send request to core services
		StatusContainer nativeStatusContainer = super.getActiveServices(companyWorkspace.getCompanyId(), serviceCode);


		// ---------------------- parse response data
		if (nativeStatusContainer != null) {
			statusContainer.setCompanyServices(parseServices(companyDispatcher,
															 nativeStatusContainer.getCompanyServices()));
			statusContainer.setApplicationServices(parseServices(applicationDispatcher,
																 nativeStatusContainer.getApplicationServices()));
		}
		
		// ---------------------- return
		return statusContainer;
	}


	/**
	 * All active services registerred at core services service.
	 *
	 * @param pWorkspace	session workspace
	 * @return				services status container
	 */
	public ServicesStatusContainer getActiveServices(DciWorkspace pWorkspace) {
		return getActiveServices(pWorkspace, null);
	}

	/**
	 * All active services registerred at core services service.
	 *
	 * @param pWorkspace	session workspace
	 * @return				services status container
	 */
	public ServicesStatusContainer getActiveService(DciWorkspace pWorkspace, String serviceCode) {
		return getActiveServices(pWorkspace, serviceCode);
	}


	private List<Service> parseServices(SystemServicesDispatcher pServiceDispatcher, List<ServiceStatus> pServices) {
		List<Service> services = new ArrayList<Service>();
		JSONObject jsonService;
		JSONObject jsonEvent;
		Service service;
		ServiceEvent event;
		SubscriptionItem opcSubscriptionItem;
		DateTime lastExecutionTime;
		
		for (ServiceStatus serviceStatus : pServices) {
			// -------------------- service modification
			service = pServiceDispatcher.getService(serviceStatus.getServiceCode());
			service.setPaused(serviceStatus.getPaused());
			service.setInterval(serviceStatus.getInterval());
			service.setActive(serviceStatus.getActive());
			service.setServiceCheckConnectionTimeout(Convert.toString(serviceStatus.getServiceCheckTimeout()));
			service.setLastTimeScheduled(DateTime.convertToDateTime(serviceStatus.getLastTimeScheduled()));

			lastExecutionTime = serviceStatus.getLastTime() != null ?
					DateTime.convertToDateTime(serviceStatus.getLastTime()) :
					DateTime.NULL_DATE;
			service.setLastTime(service.getLastTime().after(lastExecutionTime) ? service.getLastTime() : lastExecutionTime);
			
			// -------------------- events modification
			for (ServiceEventStatus eventStatus : serviceStatus.getEvents()) {
				event = service.findEvent(eventStatus.getCode());
				
				if (event != null) {
					event.setScheduled(eventStatus.getScheduled());
					event.setPeriod(eventStatus.getPeriod());
					event.setTimeUnit(eventStatus.getTimeUnit());
					event.setPeriodPattern(eventStatus.getPeriodPattern());
					event.setEventState(ServiceEventState.valueOf(eventStatus.getState().getValue()));

					lastExecutionTime = eventStatus.getLastExecution() != null ?
							DateTime.convertToDateTime(eventStatus.getLastExecution()) :
							DateTime.NULL_DATE;
					event.setLastExecution(event.getLastExecution().after(lastExecutionTime) ? event.getLastExecution() : lastExecutionTime);
					
					// ------------------------ opc subscriptions
					event.setOpcSubscriptions(new ArrayList<SubscriptionItem>());

					for (OPCSubscriptionItemStatus opcSubscriptionItemStatus : eventStatus.getOpcSubscriptions()) {
						opcSubscriptionItem = new SubscriptionItem(
								getJsonConverter().convertValue(opcSubscriptionItemStatus, cz.aimtec.dci.coregate.api.model.coreservices.OPCSubscriptionItemStatus.class)
						);
						event.getOpcSubscriptions().add(opcSubscriptionItem);
					}
				}
			}
			
			// -------------------- service to service list
			services.add(service);
		}
		
		// ----------------------- return
		return services;
	}

	/**
	 * Service restart.
	 *
	 * @param pService		reinitialized service instance (contains new interval f.e.)
	 * @return				flag indicating the action was successful
	 */
	public boolean restartService(Service pService) {
		return super.restartService(pService.createServiceConfiguration(getJsonConverter()));
	}
	
	/**
	 * Service event restart.
	 *
	 * @param pService		reinitialized service instance (contains new event parameters - period f.e.)
	 * @param pEventCode	restarted event code
	 */
	public void restartEvent(Service pService, String pEventCode) {
		super.restartEvent(pService.createServiceConfiguration(getJsonConverter()), pEventCode);
	}
	
	/**
	 * Service event command sending.
	 *
	 * @param pServiceCode		service code
	 * @param pCompanyId		company identifier which the service belongs to (null for application services)
	 * @param pEventCode		event code
	 * @param pCommandName		command name
	 * @param pCommandData		command additional data
	 * @param pMachineEvent		command machine event
	 * @param pResponseTimeout	response timeout in milliseconds
	 * @return					response message container
	 */
	public CommandMessageResponse sendEventCommand(String pServiceCode,
	                                               Integer pCompanyId,
	                                               String pEventCode,
	                                               String pCommandName,
	                                               List<String> pCommandData,
	                                               String pMachineEvent,
	                                               Integer pResponseTimeout) {
		// ----------------------------- send request to core services
		CommandMessage command = createCommandMessage(pServiceCode,
		                                            pCompanyId,
		                                            pEventCode,
		                                            pCommandName,
		                                            pCommandData,
		                                            pMachineEvent,
		                                            pResponseTimeout);
		
		CommandMessageResponse commandResponse = super.sendEventCommand(command);
		
		// ----------------------------- return
		return commandResponse;
	}
	
	/**
	 * Service event command response reading.
	 *
	 * @param pServiceCode		service code
	 * @param pCompanyId		company identifier which the service belongs to (null for application services)
	 * @param pEventCode		event code
	 * @param pCommandName		command name
	 * @param pCommandData		command additional data
	 * @param pMachineEvent		command machine event
	 * @return					response message container
	 */
	public CommandMessageResponse readEventCommandResponse(String pServiceCode,
	                                                       Integer pCompanyId,
	                                                       String pEventCode,
	                                                       String pCommandName,
	                                                       List<String> pCommandData,
	                                                       String pMachineEvent) {
		// ----------------------------- send request to core services
		CommandMessage command = createCommandMessage(pServiceCode,
		                                            pCompanyId,
		                                            pEventCode,
		                                            pCommandName,
		                                            pCommandData,
		                                            pMachineEvent,
		                                            null);
		
		CommandMessageResponse commandResponse = super.readEventCommandResponse(command);
		
		// ----------------------------- return
		return commandResponse;
	}
	
	/**
	 * Service event command response checking.
	 *
	 * @param pServiceCode		service code
	 * @param pCompanyId		company identifier which the service belongs to (null for application services)
	 * @param pEventCode		event code
	 * @param pCommandName		command name
	 * @param pCommandData		command additional data
	 * @param pMachineEvent		command machine event
	 * @return					flag indicating existing command response
	 */
	public boolean hasEventCommandResponse(String pServiceCode,
	                                       Integer pCompanyId,
	                                       String pEventCode,
	                                       String pCommandName,
	                                       List<String> pCommandData,
	                                       String pMachineEvent) {
		
		// ----------------------------- send request to core services
		CommandMessage command = createCommandMessage(pServiceCode,
		                                            pCompanyId,
		                                            pEventCode,
		                                            pCommandName,
		                                            pCommandData,
		                                            pMachineEvent,
		                                            null);
		boolean response = super.hasEventCommandResponse(command);
		
		// ----------------------------- return
		return response;
	}
	
	/**
	 * Method provides all service applicable command names.
	 * It is implemented for TCPIP listener only.
	 *
	 * @param pServiceCode	service code
	 * @param pCompanyId	company identifier which the service belongs to (null for application services)
	 * @return				list of applicable commands names
	 */
	public List<String> getServiceAllowedCommandNames(String pServiceCode,
	                                                  Integer pCompanyId) {
		return super.getServiceAllowedCommandNames(pServiceCode, pCompanyId);
	}
	
	/**
	 * Method provides all service watchdog type names.
	 * It is implemented for TCPIP listener only.
	 *
	 * @return				list of watchdog type names
	 */
	public List<String> getServiceWatchdogTypes() {
		return super.getServiceWatchdogTypes();
	}
	
	/**
	 * Service event OPC command sending.
	 *
	 * @param pServiceCode		service code
	 * @param pCompanyId		company identifier which the service belongs to (null for application services)
	 * @param pEventCode		event code
	 * @param pCommandName		command name
	 * @param pCommandData		command additional data
	 * @return					response message container
	 * @throws ApplicationException 
	 */
	public CommandMessageResponse sendOpcCommand(String pServiceCode,
	                                             Integer pCompanyId,
	                                             String pEventCode,
	                                             String pCommandName,
	                                             List<String> pCommandData) throws ApplicationException {
		// ----------------------------- send request to core services
		CommandMessage command =
				createCommandMessage(pServiceCode, pCompanyId, pEventCode, pCommandName, pCommandData, null, null);
		CommandMessageResponse commandResponse = super.sendOpcCommand(command);
			
		// ----------------------------- return
		return commandResponse;
	}
	
	/**
	 * Method reads OPC node hierarchy.
	 *
	 * @param pServiceCode		service code
	 * @param pCompanyId		company identifier which the service belongs to (null for application services)
	 * @param pEventCode		event code
	 * @param pNodeId			parent OPC node identifier
	 * @return
	 * @throws ApplicationException
	 */
	public List<OPCNodeHierarchy> readOpcNodeHierarchy(String pServiceCode,
	                                                   Integer pCompanyId,
	                                                   String pEventCode,
	                                                   String pNodeId) throws ApplicationException {
		// ----------------------------- send request to core services
		List<String> commandData = Arrays.asList(new String[]{pNodeId}); 
		CommandMessage command =
				createCommandMessage(pServiceCode, pCompanyId, pEventCode, null, commandData, null, null);
		List<OPCNodeHierarchy> nodeHierarchy = super.readOpcNodeHierarchy(command);

		// ----------------------------- return
		return nodeHierarchy;
	}
	
	/**
	 * Service event OPC subscriptions refreshment.
	 *
	 * @param pService		service instance (usually TCP/IP Listener)
	 */
	public void refreshOpcSubscriptions(Service pService) throws ApplicationException {
		super.refreshOpcSubscriptions(pService.createServiceConfiguration(getJsonConverter()));
	}

	private CommandMessage createCommandMessage(String pServiceCode,
												Integer pCompanyId,
												String pEventCode,
												String pCommandName,
												List<String> pCommandData,
												String pMachineEvent,
												Integer pResponseTimeout) {
		CommandMessage command = new CommandMessage();
		command.setServiceCode(pServiceCode);
		command.setCompanyId(pCompanyId);
		command.setEventCode(pEventCode);
		command.setCommandName(pCommandName);
		command.setCommandData(Convert.isNotBlank(pCommandData) ? pCommandData : new ArrayList<String>());
		command.setMachineEvent(pMachineEvent);
		command.setResponseTimeout(pResponseTimeout);
		
		// ---------------------------- return
		return command;
	}
	
	/**
	 * Get direct children nodes for given rootNode of request in FancyTree node format
	 * that is used for rendering result
	 *
	 * @return Available nodes or error code
	 */
	public BrowseOPCNodesResult browseOpcNodes(String pServiceCode,
													 Integer pCompanyId,
													 String pOpcServerCode,
													 FancyTreeNode pBrowseRoot,
													 String pTarget,
													 List<String> pPathToTarget) {
		
		BrowseOPCNodesRequest pRequest = new BrowseOPCNodesRequest()
				.serviceCode(pServiceCode)
				.companyId(pCompanyId)
				.opcServerCode(pOpcServerCode)
				.rootNode(pBrowseRoot)
				.target(pTarget)
				.pathToTarget(pPathToTarget);

		// ---------------------- return
		return super.browseOpcNodes(pRequest);
	}
}
