package cz.aimtec.dci.services;

import cz.aimtec.dci.DciApplicationServlet;
import cz.aimtec.dci.DciApplicationWorkspace;
import cz.aimtec.util.Convert;
import cz.aimtec.util.config.ReloadableConfigProperties;
import cz.aimtec.util.config.ReloadableConfigPropertiesNotification;
import cz.aimtec.util.config.ReloadableConfigPropertiesObserver;
import cz.aimtec.util.config.ReloadableConfigPropertiesSettings;

import java.util.Arrays;
import java.util.List;

/**
 * CoreServices system services monitor
 *
 * periodically reads changes in state of services and then sends info about those to Core
 * using EventBus
 *
 * reloadable from ApplicationProperties
 * keys:
 * 	SystemServicesMonitor.enabled
 * 	SystemServicesMonitor.interval
 *
 *
 * @author koup
 */
public class SystemServicesMonitor extends Service implements ReloadableConfigProperties {

	private static final String OBSERVED_KEY_SYSTEM_SERVICES_MONITOR_ENABLED = "SystemServicesMonitor.enabled";
	private static final String OBSERVED_KEY_SYSTEM_SERVICES_MONITOR_INTERVAL = "SystemServicesMonitor.interval";

	private static final List<String> OBSERVED_KEYS = Arrays.asList(OBSERVED_KEY_SYSTEM_SERVICES_MONITOR_ENABLED, OBSERVED_KEY_SYSTEM_SERVICES_MONITOR_INTERVAL);

	private final ReloadableConfigPropertiesSettings settings = new ReloadableConfigPropertiesSettings(OBSERVED_KEYS);

	// Reloadable configuration settings
	public static final String RELOADABLE_CODE = "SystemServicesMonitor";

	private static final int DEFAULT_POLLING_INTERVAL = 5000;

	/**
	 * Constructs a service.
	 * The service immediately creates its own thread. Be careful and create
	 * the service only if you are absolutely sure that the service should run.
	 *
	 * @param serviceElement
	 *            A service XML element.
	 */
	public SystemServicesMonitor(ServiceXmlElement serviceElement) {
		super(serviceElement);

		// if interval is not set in config, use default interval
		if (getInterval() <= 0) {
			setInterval(DEFAULT_POLLING_INTERVAL);
		}

		ReloadableConfigPropertiesObserver.getInstance().register(RELOADABLE_CODE, this, settings);
	}

	@Override
	protected void execute() {
		//DO NOTHING
	}

	@Override
	public void reload(List<ReloadableConfigPropertiesNotification> notifications) {
		DciApplicationWorkspace applicationWorkspace = DciApplicationWorkspace.getInstance();
		DciApplicationServlet applicationServlet = (DciApplicationServlet) applicationWorkspace.getApplicationServlet();
		SystemServicesDispatcher systemServicesDispatcher = applicationServlet.getSystemServicesDispatcher();

		for (ReloadableConfigPropertiesNotification notification: notifications) {
			if (OBSERVED_KEY_SYSTEM_SERVICES_MONITOR_INTERVAL.equalsIgnoreCase(notification.getKey())) {
				this.setInterval(Convert.toLong(notification.getNewValue()));
				if(isEnabled()) {
					systemServicesDispatcher.restartService(RELOADABLE_CODE, this.getInterval());
				}
			}
			if (OBSERVED_KEY_SYSTEM_SERVICES_MONITOR_ENABLED.equalsIgnoreCase(notification.getKey())) {
				setEnabled(Convert.stringToSmallBoolean(notification.getNewValue()));
				systemServicesDispatcher.restartService(RELOADABLE_CODE);
			}
		}
	}
}
