package cz.aimtec.dci.websocket.server;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

import cz.aimtec.dci.user.User;
import cz.aimtec.util.Convert;


/**
 * Manager for Web Socket connections. It provides an API for releasing connections based on unique codes.
 * These codes can be shared between multiple clients.
 * 
 * @author smem, bejd
 *
 */
public class WebSocketManager {
	
	private static final Logger log = Logger.getLogger(WebSocketManager.class);
	
	// Two different maps are used, because we don't want to use three-dimension maps
	private static HashMap<String, List<WebSocketServer>> connectionsSemaphoreMap = new HashMap<>();
	private static HashMap<String, List<WebSocketServer>> connectionsMessagesMap = new HashMap<>();
	private static HashMap<String, List<WebSocketServer>> connectionsSystemServicesMap = new HashMap<>();
	
	private static final String RELEASE_COMMAND = "release";
	public static final String PING_MESSAGE = "ping";
	public static final String PONG_MESSAGE = "pong";
	
	private WebSocketManager() {
		// Utility class
	}
	
	/**
	 * Register a new Web Socket server.
	 * 
	 * @param server
	 */
	public static void register(final WebSocketServer server) {
		WebSocketType socketType = server.getSocketType();
		Map<String, List<WebSocketServer>> sourceMap = socketType.chooseSourceMap();
		String code = socketType.getConnectionIdentifier(server);
		
		if (log.isDebugEnabled()) {
			log.debug("Websocket register with code: " + code);
		}
		
		if (!sourceMap.containsKey(code)) {
			List<WebSocketServer> servers = new ArrayList<>();
			sourceMap.put(code, servers);
		}
		sourceMap.get(code).add(server);
		
		if (log.isDebugEnabled()) {
			log.debug("Websocket register done with code: " + code + ", map size: " + Convert.toString(sourceMap.get(code).size()));
		}
	}
	
	/**
	 * Release existing connection.
	 * @param type
	 * @param connectionIdentifier identifier for connection in connections map, 
	 * 							   is specified by method WebSocketType.getConnectionIdentifier()
	 */
	public static void release(final WebSocketType type, final String connectionIdentifier) {
		Map<String, List<WebSocketServer>> sourceMap = type.chooseSourceMap();
		List<WebSocketServer> connections = sourceMap.get(connectionIdentifier);

		if (connections != null) {
			List<WebSocketServer> clonedConnections = new ArrayList<>(connections);
			for (int i = 0; i < clonedConnections.size(); i++) {
				WebSocketServer server = clonedConnections.get(i);
				if (server.getSession().isOpen() && server.isConnected()) {
					server.sendMessage(RELEASE_COMMAND);
				} else {
					unregister(server, false);
				}
			}
		}
	}
	
	/**
	 * Release all connections whose code (in the format "code[subcode1,subcode2,subcode3]") correspond to the code, subCode parameters. 
	 * Example: connection with the code "red[alpha, beta]" is released for the following calls:
	 * 		release("red", "alpha");
	 * 		release("red", "beta"); 
	 * 
	 * @param code
	 * @param subCode
	 */
	public static void release(final WebSocketType type, final String code, final String subCode) {		
		Map<String, List<WebSocketServer>> sourceMap = type.chooseSourceMap();
		Set<String> keys = sourceMap.keySet();
		keys.forEach(connectionIdentifier -> {
			if (containsCode(connectionIdentifier, code) && containsSubcode(connectionIdentifier, subCode)) {
				release(type, connectionIdentifier);
			}
		});
	}

	public static void reconnect(final WebSocketServer server) {
		Map<String, List<WebSocketServer>> sourceMap = server.getSocketType().chooseSourceMap();
		List<WebSocketServer> connections = sourceMap.get(server.getUniqueCode());

		if (log.isDebugEnabled()) {
			log.debug("Websocket reconnect with userId: " + server.getUserId() 
					+ ", sessionId: " + server.getSessionId() 
					+ ", connections size: " + (connections == null ? "empty" : connections.size()));
		}
		
		// Check case if semaphore after reconnect was NOT released
		if (connections != null) {
			for (WebSocketServer webSocketServer : connections) {
				if (webSocketServer.getUserId() == server.getUserId() && webSocketServer.getSessionId().equals(server.getSessionId())) {
					unregister(webSocketServer, true);
					register(server);
					
					if (log.isDebugEnabled()) {
						log.debug("Websocket reconnected, connection size: " + connections.size());
					}
					
					return;
				}
			}
		}
		
		if (WebSocketType.SEMAPHORE.equals(server.getSocketType())) {
			// No semaphore for user and session not found => close ping thread and send release
			server.getPingThread().stopThread();
			server.sendMessage(RELEASE_COMMAND);
		} else {
			// other ws types should be registered again
			register(server);
			
			if (log.isDebugEnabled()) {
				log.debug("Websocket registered with userId: " + server.getUserId() + ", connections size: " + (connections != null ? connections.size() : "empty"));
			}
		}
	}
	
	private static boolean containsCode(String connectionIdentifier, String pCode) {		
		int end = connectionIdentifier.indexOf('[');
		String code = connectionIdentifier.substring(0, end);
		return pCode.equals(code);
	}
	
	protected static boolean containsSubcode(String connectionIdentifier, String subCode) {
		List<String> list = Convert.stringValuesInBracketsToList(
				connectionIdentifier.substring(
						connectionIdentifier.indexOf('[')));
		return list.contains(subCode);		
	}

	/**
	 * Unregister Web Socket Server. Note that the connection is not 'released' (clients won't get any
	 * 		reason why it was closed).
	 * @param server
	 */
	public static void unregister(final WebSocketServer server) {
		unregister(server, false);
	}
	
	/**
	 * Unregister Web Socket Server. Note that the connection is not 'released' (clients won't get any
	 * 		reason why it was closed).
	 * @param server
	 * @param alreadyClosed unregister already closed connection. Do not attempt to close it again.
	 */
	public static void unregister(final WebSocketServer server, final boolean alreadyClosed) {
		WebSocketType socketType = server.getSocketType();
		if (socketType == null) {
			return;
		}
		Map<String, List<WebSocketServer>> sourceMap = socketType.chooseSourceMap();
		if (!alreadyClosed && server.getSession().isOpen()) {
			server.onClose();
		}
		List<WebSocketServer> servers = sourceMap.get(socketType.getConnectionIdentifier(server)); 
		if (servers != null) {
			servers.remove(server);
			if (servers.isEmpty()) {
				sourceMap.remove(socketType.getConnectionIdentifier(server));
			}
			
			if (log.isDebugEnabled()) {
				log.debug("Websocket unregistered with userId: " + server.getUserId() + ", servers size: " + (servers != null ? servers.size() : "empty"));
			}
		}
		
		if (log.isDebugEnabled()) {
			log.debug("Websocket unregister done with userId:" + server.getUserId() + ", servers size: " + (servers != null ? servers.size() : "empty"));
		}
	}
	
	/**
	 * Unregister Web Socket Server. Note that the connection is not 'released' (clients won't get any
	 * 		reason why it was closed).
	 * @param type
	 * @param connectionIdentifier identifier for connection in connections map, 
	 * 							   is specified by method WebSocketType.getConnectionIdentifier()
	 */
	public static void unregister(final WebSocketType type, final String connectionIdentifier){
		unregister(type, connectionIdentifier, false);
	}
	
	/**
	 * Unregister Web Socket Server. Note that the connection is not 'released' (clients won't get any
	 * 		reason why it was closed).
	 * @param type
	 * @param connectionIdentifier identifier for connection in connections map, 
	 * 							   is specified by method WebSocketType.getConnectionIdentifier()
	 * @param alreadyClosed unregister already closed connection. Do not attempt to close it again.
	 */
	public static void unregister(final WebSocketType type, final String connectionIdentifier, final boolean alreadyClosed) {
		Map<String, List<WebSocketServer>> sourceMap = type.chooseSourceMap();
		List<WebSocketServer> connections = sourceMap.get(connectionIdentifier);
		
		if (log.isDebugEnabled()) {
			log.debug("Websocket unregister with identifier: " + connectionIdentifier + " , map size: " + (sourceMap != null ? sourceMap.size() : "empty"));
		}
		
		if (connections != null) {
			if (!alreadyClosed) {
				connections.forEach(server -> { 
					if (server.getSession().isOpen()) {
						server.onClose();
					}
				});
			}
			sourceMap.remove(connectionIdentifier);
		}
		
		if (log.isDebugEnabled()) {
			log.debug("Websocket unregister with identifier: " + connectionIdentifier + " done, map size: " + (sourceMap != null ? sourceMap.size() : "empty"));
		}
	}
	
	/**
	 * Unregister all existing connection for specified user.
	 * @param type
	 * @param user
	 */
	public static void unregister(final WebSocketType type, final User user) {
		Map<String, List<WebSocketServer>> sourceMap = type.chooseSourceMap();
		
		if (log.isDebugEnabled()) {
			log.debug("Websocket unregister, map size: " + (sourceMap != null ? sourceMap.size() : "empty"));
		}
		
		for (String connectionIdentifier : sourceMap.keySet()) {
			List<WebSocketServer> connections = sourceMap.get(connectionIdentifier);
			if (connections != null) {
				connections.removeIf(entry -> entry.getUserId().equals(user.getId()));
			}
		}
		sourceMap.keySet().removeIf(entry -> sourceMap.get(entry).isEmpty());
		
		if (log.isDebugEnabled()) {
			log.debug("Websocket unregister done, map size: " + (sourceMap != null ? sourceMap.size() : "empty"));
		}
	}
	
	public static void onLostConnection(WebSocketServer server) {
		server.setIsConnected(false);
		log.warn("Lost connection with client.");
	}
	
	/**********************************************************
	 * Friend design pattern. These methods are accessible only by friends.
	 * Friends in Java = Classes in the same packages.
	 * 
	 * This gives us the possibility to expose maps for WebSocketType enum, but restrict it for anyone else, because
	 * we don't really want to give access to these maps.
	 *********************************************************/
	
	static Map<String, List<WebSocketServer>> getConnectionsSemaphoreMap() {
		return connectionsSemaphoreMap;
	}
	
	static Map<String, List<WebSocketServer>> getConnectionsMessagesMap() {
		return connectionsMessagesMap;
	}

	static Map<String, List<WebSocketServer>> getSystemServicesConnectionMap() {
		return connectionsSystemServicesMap;
	}
	
	/****************************************************/

}
