package cz.aimtec.dci.coregate.coreservices;

import com.fasterxml.jackson.core.JsonProcessingException;
import cz.aimtec.dci.coregate.api.model.coreservices.*;
import cz.aimtec.dci.microservice.client.HttpMethod;
import cz.aimtec.dci.microservice.client.ServiceClientAsync;
import cz.aimtec.dci.microservice.client.ServiceResponse;
import cz.aimtec.dci.microservice.discovery.DiscoveryService;
import cz.aimtec.dci.microservice.eventbus.ClientResponseHandler;
import cz.aimtec.dci.microservice.eventbus.EventBusClient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.HttpStatus;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Core services part consumption client.
 */
public class ServiceConsumptionClient extends ServiceClientAsync {
	protected final EventBusClient eventBusClient;

    private static final String COMPANY_ID_PARAMETER_NAME = "companyId";
    private static final String SERVICE_CODE_PARAMETER_NAME = "serviceCode";

    private final Logger logger = LoggerFactory.getLogger(this.getClass());

    /**
     * Basic constructor.
     *
     * @param pDiscovery       service discovery provider
     * @param pEventBusClient  event bus client
     * @param pServiceName     service name
     * @param pRequiredVersion service version
     */
    public ServiceConsumptionClient(DiscoveryService pDiscovery,
                                    EventBusClient pEventBusClient,
                                    String pServiceName,
                                    String pRequiredVersion) {
        super(pDiscovery, pEventBusClient, pServiceName, pRequiredVersion);
        this.eventBusClient = pEventBusClient;
	}

	/**
	 * Basic constructor.
	 *
	 * @param pDiscovery       service discovery provider
	 * @param pEventBusClient  event bus client
	 * @param pServiceName     service name
	 */
	public ServiceConsumptionClient(DiscoveryService pDiscovery,
									EventBusClient pEventBusClient,
									String pServiceName) {
		super(pDiscovery, pEventBusClient, pServiceName);
		this.eventBusClient = pEventBusClient;
	}

	/**
	 * Basic contructore
	 */
	public ServiceConsumptionClient() {
		super();
		this.eventBusClient = null;
	}


	@Override
	protected String getPomPropertiesFilePath() {
		return "cz/aimtec/dci/coregate/";
	}

	/**
	 * Method provides all core services to be activated.
	 *
	 * @return	services for activation or null (if it is not possible to get the list)
	 */
	public List<ServiceConfiguration> getServicesConfiguration() throws JsonProcessingException {
		List<ServiceConfiguration> servicesConfiguration = null;

		// ---------------------- send request to core gate
		ServiceResponse response = callMethod(HttpMethod.GET, "coreServices/configuration");

		// ---------------------- parse response data
		if (response.getStatus() == (HttpStatus.OK.value())) {
			ServiceConfiguration[] servicesConfigurationNative = getJsonConverter().readValue(response.getData(),
																							  ServiceConfiguration[].class);
			servicesConfiguration = Arrays.asList(servicesConfigurationNative);
		}

		// ---------------------- return
		return servicesConfiguration;
	}

	/**
	 * Method executes core service.
	 *
	 * @param pServiceCode	service code
	 * @param pCompanyId	identifier of company containing the service
	 */
	public void executeService(String pServiceCode, Integer pCompanyId) {
		executeServiceOrEvent(pServiceCode, pCompanyId, null);
	}

	/**
	 * Method executes core service event.
	 *
	 * @param pServiceCode	service code
	 * @param pCompanyId	identifier of company containing the service
	 * @param pEventCode	service event code
	 * @return				flag indicating success of execution
	 */
	public boolean executeServiceEvent(String pServiceCode, Integer pCompanyId, String pEventCode) {
		return executeServiceOrEvent(pServiceCode, pCompanyId, pEventCode);
	}

	private boolean executeServiceOrEvent(String pServiceCode, Integer pCompanyId, String pEventCode) {
		// ---------------------- service execution directive completion
		ServiceExecution serviceExecution = new ServiceExecution();
		serviceExecution.setServiceCode(pServiceCode);
		serviceExecution.setCompanyId(pCompanyId);
		serviceExecution.setEventCode(pEventCode);

		// ---------------------- send request to core gate
		ServiceResponse response = callMethod(HttpMethod.POST, "coreServices/execute", serviceExecution);

		// ---------------------- response handling
		if (response.getStatus() != (HttpStatus.OK.value())) {
			logger.error("Can not execute service. [" + pServiceCode + ", " + pCompanyId + ", " + pEventCode +
								 "]. HTTP status: " + response.getStatus());
			return false;
		}

		// ---------------------- positive return
		return true;
	}

	/**
	 * Method registers core service resource usage license.
	 *
	 * @param pServiceCode	service code
	 * @param pCompanyId	identifier of company containing the service
	 * @param pEventCode	service event code
	 */
	public void registerLicense(String pServiceCode, Integer pCompanyId, String pEventCode) {
		// ---------------------- service execution directive completion
		ServiceExecution serviceExecution = new ServiceExecution();
		serviceExecution.setServiceCode(pServiceCode);
		serviceExecution.setCompanyId(pCompanyId);
		serviceExecution.setEventCode(pEventCode);

		// ---------------------- send request to core gate
		ServiceResponse response = callMethod(HttpMethod.POST, "coreServices/registerLicense", serviceExecution);

		// ---------------------- response handling
		if (response.getStatus() != (HttpStatus.OK.value())) {
			logger.error("Can not register service resource usage license. [" +
								 pServiceCode + ", " + pCompanyId + ", " + pEventCode + "]. HTTP status: " + response.getStatus());
		}
	}

	/**
	 * Method deregisters core service resource usage license.
	 *
	 * @param pServiceCode	service code
	 * @param pCompanyId	identifier of company containing the service
	 * @param pEventCode	service event code
	 */
	public void deregisterLicense(String pServiceCode, Integer pCompanyId, String pEventCode) {
		// ---------------------- service execution directive completion
		ServiceExecution serviceExecution = new ServiceExecution();
		serviceExecution.setServiceCode(pServiceCode);
		serviceExecution.setCompanyId(pCompanyId);
		serviceExecution.setEventCode(pEventCode);

		// ---------------------- send request to core gate
		ServiceResponse response = callMethod(HttpMethod.POST, "coreServices/deregisterLicense", serviceExecution);

		// ---------------------- response handling
		if (response.getStatus() != (HttpStatus.OK.value())) {
			logger.error("Can not deregister service resource usage license. [" +
								 pServiceCode + ", " + pCompanyId + ", " + pEventCode + "]. HTTP status: " + response.getStatus());
		}
	}

	/**
	 * Method handles incoming listener message.
	 *
	 * @param pMessage		incoming message to be handled
	 */
	public void handleMessage(IncomingMessage pMessage) {
		String serviceCode = pMessage.getServiceCode();
		Integer companyId = pMessage.getCompanyId();
		String eventCode = pMessage.getEventCode();

		// ---------------------- send request to core gate
		ServiceResponse response = callMethod(HttpMethod.POST,"coreServices/handleMessage", pMessage);

		// ---------------------- response handling
		if (response.getStatus() != (HttpStatus.OK.value())) {
			logger.error("Can not handle incoming message. [" +
								 serviceCode + ", " + companyId + ", " + eventCode + "]. HTTP status: " + response.getStatus());
		}
	}

	/**
	 * Messages archiving.
	 * It moves active incoming event messages into archive table.
	 *
	 * @param pMessages	list of messages to be archived
	 */
	public void archiveMessages(List<IncomingMessage> pMessages) {
		// ---------------------- send request to core gate
		ServiceResponse response = callMethod(HttpMethod.POST, "coreServices/archiveMessages", pMessages);

		// ---------------------- response handling
		if (response.getStatus() != (HttpStatus.OK.value())) {
			logger.error("Can not archive incoming messages. HTTP status: " + response.getStatus());
		}
	}

	/**
	 * Watchdog logs saving.
	 *
	 * @param pServiceCode	service code
	 * @param pCompanyId	identifier of company containing the service
	 * @param pWatchdogLogs	watchdog logs to be saved
	 */
	public void saveWatchdogLogs(String pServiceCode,
								 Integer pCompanyId,
								 List<WatchdogLog> pWatchdogLogs) {
		// ---------------------- query parameters completion
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);

		// ---------------------- send request to core gate
		ServiceResponse response = callMethod(HttpMethod.POST, "coreServices/saveWatchdogLogs", pWatchdogLogs, queryParameters);

		// ---------------------- response handling
		if (response.getStatus() != (HttpStatus.OK.value())) {
			logger.error("Can not save watchdog logs. HTTP status: " + response.getStatus());
		}
	}

	/**
	 * Service event machine initialization.
	 *
	 * @param pServiceCode	service code
	 * @param pCompanyId	identifier of company containing the service
	 * @param pEventCode	service event code
	 */
	public void initializeMachine(String pServiceCode, Integer pCompanyId, String pEventCode) {
		// ---------------------- service execution directive completion
		ServiceExecution serviceExecution = new ServiceExecution();
		serviceExecution.setServiceCode(pServiceCode);
		serviceExecution.setCompanyId(pCompanyId);
		serviceExecution.setEventCode(pEventCode);

		// ---------------------- send request to core gate
		ServiceResponse response = callMethod(HttpMethod.POST, "coreServices/initializeMachine", serviceExecution);

		// ---------------------- response handling
		if (response.getStatus() != (HttpStatus.OK.value())) {
			logger.error("Can not initialize service event machine. [" +
								 pServiceCode + ", " + pCompanyId + ", " + pEventCode + "]. HTTP status: " + response.getStatus());
		}
	}

	/**
	 * SAP RFC server state saving.
	 *
	 * @param pServiceCode	service code
	 * @param pCompanyId	identifier of company containing the service
	 * @param pServerState	SAP RFC server state
	 */
	public void saveRfcServerState(String pServiceCode,
								   Integer pCompanyId,
								   SAPRfcServerState pServerState) {
		// ---------------------- query parameters completion
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);

		// ---------------------- send request to core gate
		ServiceResponse response = callMethod(HttpMethod.POST, "coreServices/saveRfcServerState", pServerState, queryParameters);

		// ---------------------- response handling
		if (response.getStatus() != (HttpStatus.OK.value())) {
			logger.error("Can not save RFC server state. HTTP status: " + response.getStatus());
		}
	}

	/**
	 * SAP RFC TID checking.
	 *
	 * @param pServiceCode	service code
	 * @param pCompanyId	identifier of company containing the service
	 * @return				SAP RFC TID status container
	 */
	public SAPRfcTidStatus checkRfcTid(String pServiceCode,
									   Integer pCompanyId) {
		SAPRfcTidStatus status = new SAPRfcTidStatus();
		status.setStatus(false);

		// ---------------------- query parameters completion
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);

		// ---------------------- send request to core gate
		ServiceResponse response = callMethod(HttpMethod.GET, "coreServices/checkRfcTid", queryParameters);

		// ---------------------- response handling
		if (response.getStatus() == (HttpStatus.OK.value())) {
			try {
				status = getJsonConverter().readValue(response.getData(), SAPRfcTidStatus.class);

			} catch (JsonProcessingException exception) {
				logger.error("Can not parse SAP TID status.", exception);
			}

		} else {
			logger.error("Can not check RFC TID. HTTP status: " + response.getStatus());
		}

		// ---------------------- return
		return status;
	}


	/**
	 * SAP RFC incoming document message handling.
	 *
	 * @param pServiceCode	service code
	 * @param pCompanyId	identifier of company containing the service
	 * @param pMessage		SAP RFC incoming document message
	 */
	public void handleRfcMessage(String pServiceCode,
								 Integer pCompanyId,
								 SAPRfcMessage pMessage) {
		// ---------------------- query parameters completion
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);

		// ---------------------- send request to core gate
		ServiceResponse response = callMethod(HttpMethod.POST, "coreServices/handleRfcMessage", pMessage, queryParameters);

		// ---------------------- response handling
		if (response.getStatus() != (HttpStatus.OK.value())) {
			logger.error("Can not handle RFC incoming message. HTTP status: " + response.getStatus());
		}
	}

	/**
	 * SAP RFC server error saving.
	 *
	 * @param pServiceCode	service code
	 * @param pCompanyId	identifier of company containing the service
	 * @param pServerError	SAP RFC server error
	 */
	public void saveRfcServerError(String pServiceCode,
								   Integer pCompanyId,
								   SAPRfcServerError pServerError) {
		// ---------------------- query parameters completion
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);

		// ---------------------- send request to core gate
		ServiceResponse response = callMethod(HttpMethod.POST, "coreServices/saveRfcServerError", pServerError, queryParameters);

		// ---------------------- response handling
		if (response.getStatus() != (HttpStatus.OK.value())) {
			logger.error("Can not save RFC server error. HTTP status: " + response.getStatus());
		}
	}

	/**
	 * Machine response duration lo saving.
	 *
	 * @param pServiceCode	service code
	 * @param pCompanyId	company identifier
	 * @param pEventCode	event code
	 * @param pOpcNodeId 	OPC node id
	 * @param pMachineCode	related machine code
	 * @param pChangedValue	machine changed value
	 * @param pAuthor		author (who caused the value change)
	 */
	public void saveMachineResponseDuration(String pServiceCode,
											Integer pCompanyId,
											String pEventCode,
											String pOpcNodeId,
											String pMachineCode,
											String pChangedValue,
											String pAuthor) {
		// ---------------------- query parameters completion
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);
		queryParameters.put("eventCode", pEventCode);

		// ---------------------- send request to core gate
		MachineResponseDuration responseDuration = new MachineResponseDuration();
		responseDuration.setMachineCode(pMachineCode);
		responseDuration.setOpcNodeId(pOpcNodeId);
		responseDuration.setChangedValue(pChangedValue);
		responseDuration.setAuthor(pAuthor);

		ServiceResponse response = callMethod(HttpMethod.POST, "coreServices/saveMachineResponseDuration", responseDuration, queryParameters);

		// ---------------------- response handling
		if (response.getStatus() != (HttpStatus.OK.value())) {
			logger.error("Can not save machine response duration log record. HTTP status: " + response.getStatus());
		}
	}

	public void sendServicesStateToDCI(StatusContainer servicesContainer) {
		ClientResponseHandler responseHandler = pResponse -> {
			logger.debug("RESPONSE from EventBus:" + pResponse);
		};

		callMethodAsync(HttpMethod.POST, "coreServices/monitorSystemServices", null, servicesContainer, null, responseHandler);
	}
}
