package cz.aimtec.dci.coregate.service;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import cz.aimtec.dci.coregate.api.model.coreservices.*;
import cz.aimtec.dci.microservice.client.HttpMethod;
import cz.aimtec.dci.microservice.client.ServiceClient;
import cz.aimtec.dci.microservice.client.ServiceResponse;
import cz.aimtec.util.Convert;
import cz.aimtec.util.datatype.DateTime;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;


/**
 * DCI Core services access gate.
 *
 * @author	hors
 * @version	31. 10. 2017/hors - created<br>
 */
@Service
public class ServicesProvider extends Connector {

	private static final DateFormat dateFormat = new SimpleDateFormat(DateTime.SYSTEM_DATETIME_FORMAT);

	private static final String EVENT_CODE_PARAMETER_NAME = "eventCode";
	private static final String COMPANY_ID_PARAMETER_NAME = "companyId";
	private static final String SERVICE_CODE_PARAMETER_NAME = "serviceCode";
	private static final String ACTION_PARAMETER_NAME = "action";
	private static final String CORE_SERVICES_PATH = "/coreServices.do";
	private static final String CONFIGURATION_ACTION = "configuration";
	private static final String EXECUTE_ACTION = "execute";
	private static final String REGISTER_LICENSE_ACTION = "registerLicense";
	private static final String DEREGISTER_LICENSE_ACTION = "deregisterLicense";
	private static final String HANDLE_MESSAGE = "handleMessage";
	private static final String ARCHIVE_MESSAGES = "archiveMessages";
	private static final String SAVE_WATCHDOG_LOGS = "saveWatchdogLogs";
	private static final String INITIALIZE_MACHINE = "initializeMachine";
	private static final String SAVE_RFC_SERVER_STATE = "saveRfcServerState";
	private static final String CHECK_RFC_TID = "checkRfcTid";
	private static final String HANDLE_RFC_MESSAGE = "handleRfcMessage";
	private static final String SAVE_RFC_SERVER_ERROR = "saveRfcServerError";
	private static final String SAVE_MACHINE_RESOPONSE_DURATION = "saveMachineResponseDuration";
	private static final String MONITOR_SYSTEM_SERVICES = "servicesMonitor";

	@Autowired
	private ObjectMapper jsonConverter;


	/**
	 * Method provides all activated core services list.
	 *
	 * @return	list of activated services or null when core can not be reached
	 */
	public List<ServiceConfiguration> getConfiguration() throws JsonProcessingException {
		List<ServiceConfiguration> services = null;

		// ------------------------------- sending the request
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(ACTION_PARAMETER_NAME, CONFIGURATION_ACTION);
		ServiceResponse response = ServiceClient.sendRequest(HttpMethod.GET,
		                                                     getServer(),
		                                                     getPort(),
		                                                     getPath() + CORE_SERVICES_PATH,
		                                                     queryParameters);

		// ------------------------------- response processing
		if (HttpStatus.OK.value() == response.getStatus()) {
				ServiceConfiguration[] nativeServices = jsonConverter.readValue(response.getData(),
																				ServiceConfiguration[].class);
				services = Arrays.asList(nativeServices);
			}
		
		// ------------------------------ return
		return services;
	}
	
	/**
	 * Method executes service or service event.
	 *
	 * @param pServiceCode	executed service code
	 * @param pCompanyId	company including activated service identifier
	 * @param pEventCode	executed event code
	 * @return				flag indicating if service has been executed
	 */
	public boolean execute(String pServiceCode, Integer pCompanyId, String pEventCode) {
		// ------------------------------- sending the request
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(ACTION_PARAMETER_NAME, EXECUTE_ACTION);
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);
		queryParameters.put(EVENT_CODE_PARAMETER_NAME, pEventCode);
		ServiceResponse response = ServiceClient.sendRequest(HttpMethod.GET,
		                                                     getServer(),
		                                                     getPort(),
		                                                     getPath() + CORE_SERVICES_PATH,
		                                                     queryParameters);

		// -------------------------- return
		return HttpStatus.OK.value() == response.getStatus();
	}
	
	/**
	 * Method registers service event resource usage license.
	 *
	 * @param pServiceCode	executed service code
	 * @param pCompanyId	company including activated service identifier
	 * @param pEventCode	executed event code
	 * @return				flag indicating if service license has been registerred
	 */
	public boolean registerLicence(String pServiceCode, Integer pCompanyId, String pEventCode) {
		// ------------------------------- sending the request
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(ACTION_PARAMETER_NAME, REGISTER_LICENSE_ACTION);
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);
		queryParameters.put(EVENT_CODE_PARAMETER_NAME, pEventCode);
		ServiceResponse response = ServiceClient.sendRequest(HttpMethod.GET,
		                                                     getServer(),
		                                                     getPort(),
		                                                     getPath() + CORE_SERVICES_PATH,
		                                                     queryParameters);

		// -------------------------- return
		return HttpStatus.OK.value() == response.getStatus();
	}
	
	/**
	 * Method deregisters service event resource usage license.
	 *
	 * @param pServiceCode	executed service code
	 * @param pCompanyId	company including activated service identifier
	 * @param pEventCode	executed event code
	 * @return				flag indicating if service license has been deregisterred
	 */
	public boolean deregisterLicence(String pServiceCode, Integer pCompanyId, String pEventCode) {
		// ------------------------------- sending the request
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(ACTION_PARAMETER_NAME, DEREGISTER_LICENSE_ACTION);
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);
		queryParameters.put(EVENT_CODE_PARAMETER_NAME, pEventCode);
		ServiceResponse response = ServiceClient.sendRequest(HttpMethod.GET,
		                                                     getServer(),
		                                                     getPort(),
		                                                     getPath() + CORE_SERVICES_PATH,
		                                                     queryParameters);

		// -------------------------- return
		return HttpStatus.OK.value() == response.getStatus();
	}
	
	/**
	 * Method handles incoming listener message.
	 *
	 * @param pMessage		incoming message
	 * @return				flag indicating if service license has been registerred
	 */
	public boolean handleMessage(IncomingMessage pMessage) throws JsonProcessingException {
		// ------------------------------- sending the request
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(ACTION_PARAMETER_NAME, HANDLE_MESSAGE);
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pMessage.getServiceCode());
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pMessage.getCompanyId());
		queryParameters.put(EVENT_CODE_PARAMETER_NAME, pMessage.getEventCode());
		queryParameters.put("listenerCode", pMessage.getListenerCode());
		queryParameters.put("messageType", pMessage.getMessageType());
		queryParameters.put("messageBytes", pMessage.getMessageBytes());
		queryParameters.put("dataVector", pMessage.getDataVector());
		queryParameters.put("commandResponse", pMessage.getCommandResponse());
		queryParameters.put("machineState", pMessage.getMachineState());
		queryParameters.put("description", pMessage.getDescription());
		queryParameters.put("createDate", dateFormat.format(DateTime.convertToDateTime(pMessage.getCreateDate())));
		queryParameters.put("uniqueIdentifier", pMessage.getUniqueIdentifier());
		queryParameters.put("opcNodeId", pMessage.getOpcNodeId());
		queryParameters.put("opcAttributeTypeId", pMessage.getOpcAttributeTypeId());
		queryParameters.put("transactionDefinitionName", pMessage.getTransactionDefinitionName());
		queryParameters.put("logValues", jsonConverter.writeValueAsString(pMessage.getLogValues()));
		queryParameters.put("initialParameters", jsonConverter.writeValueAsString(pMessage.getInitialParameters()));
		queryParameters.put("machineCode", pMessage.getMachineCode());

		ServiceResponse response = ServiceClient.sendRequest(HttpMethod.FORM_DATA,
		                                                     getServer(),
		                                                     getPort(),
		                                                     getPath() + CORE_SERVICES_PATH,
		                                                     queryParameters);
		
		// -------------------------- return
		return HttpStatus.OK.value() == response.getStatus();
	}

	/**
	 * Incoming service event messages archiving.
	 * Moves existing messages from live table to archive table.
	 *
	 * @param pMessages	list of messages intended to be archived
	 * @return
	 */
	public boolean archiveMessages(List<IncomingMessage> pMessages) {
		// ------------------------------- sending the request
		Map<String, Object> queryParameters = new HashMap<>();
		List<String> messageIdentifiers = pMessages.stream().map(IncomingMessage::getUniqueIdentifier).collect(Collectors.toList());
		queryParameters.put(ACTION_PARAMETER_NAME, ARCHIVE_MESSAGES);
		queryParameters.put("archivedIdentifers", Convert.toDelimitedString(messageIdentifiers));
		ServiceResponse response = ServiceClient.sendRequest(HttpMethod.POST,
		                                                     getServer(),
		                                                     getPort(),
		                                                     getPath() + CORE_SERVICES_PATH,
		                                                     queryParameters);
		
		// -------------------------- return
		return HttpStatus.OK.value() == response.getStatus();
	}
	
	/**
	 * Service watchdog logs into database saving.
	 *
	 * @param pServiceCode	executed service code
	 * @param pCompanyId	company including activated service identifier
	 * @param pWatchdogLogs	watchdog logs collection to be saved
	 * @return				flag indicating if service watchdog logs have been saved
	 */
	public boolean saveWatchdogLogs(String pServiceCode, Integer pCompanyId, List<WatchdogLog> pWatchdogLogs) throws JsonProcessingException {
		// ------------------------------- sending the request
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(ACTION_PARAMETER_NAME, SAVE_WATCHDOG_LOGS);
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);
		queryParameters.put("serializedRequestBody", jsonConverter.writeValueAsString(pWatchdogLogs));
		ServiceResponse response = ServiceClient.sendRequest(HttpMethod.FORM_DATA,
		                                                     getServer(),
		                                                     getPort(),
		                                                     getPath() + CORE_SERVICES_PATH,
		                                                     queryParameters);
		
		// -------------------------- return
		return HttpStatus.OK.value() == response.getStatus();
	}
	
	/**
	 * Machine status initialization..
	 *
	 * @param pServiceCode	executed service code
	 * @param pCompanyId	company including activated service identifier
	 * @param pEventCode	service event code
	 * @return				flag indicating if machine status has been initialized
	 */
	public boolean initializeMachine(String pServiceCode, Integer pCompanyId, String pEventCode) {
		// ------------------------------- sending the request
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(ACTION_PARAMETER_NAME, INITIALIZE_MACHINE);
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);
		queryParameters.put(EVENT_CODE_PARAMETER_NAME, pEventCode);
		ServiceResponse response = ServiceClient.sendRequest(HttpMethod.POST,
		                                                     getServer(),
		                                                     getPort(),
		                                                     getPath() + CORE_SERVICES_PATH,
		                                                     queryParameters);
			
		// -------------------------- return
		return HttpStatus.OK.value() == response.getStatus();
	}

	/**
	 * SAP RFC service server state into database saving.
	 *
	 * @param pServiceCode	executed service code
	 * @param pCompanyId	company including activated service identifier
	 * @param pServerState	server state to be saved
	 * @return				flag indicating if service server state has been saved
	 */
	public boolean saveRfcServerState(String pServiceCode,
	                                  Integer pCompanyId,
	                                  SAPRfcServerState pServerState) throws JsonProcessingException {
		// ------------------------------- sending the request
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(ACTION_PARAMETER_NAME, SAVE_RFC_SERVER_STATE);
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);
		queryParameters.put("serializedRequestBody", jsonConverter.writeValueAsString(pServerState));
		ServiceResponse response = ServiceClient.sendRequest(HttpMethod.FORM_DATA,
		                                                     getServer(),
		                                                     getPort(),
		                                                     getPath() + CORE_SERVICES_PATH,
		                                                     queryParameters);
		
		// -------------------------- return
		return HttpStatus.OK.value() == response.getStatus();
	}
	
	/**
	 * SAP RFC service TID check.
	 *
	 * @param pServiceCode	executed service code
	 * @param pCompanyId	company including activated service identifier
	 * @return				SPA RFC TID check container
	 */
	public SAPRfcTidStatus checkRfcTid(String pServiceCode,
									   Integer pCompanyId) throws JsonProcessingException {
		SAPRfcTidStatus status = new SAPRfcTidStatus();
		status.setStatus(false);
		
		// ------------------------------- sending the request
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(ACTION_PARAMETER_NAME, CHECK_RFC_TID);
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);
		ServiceResponse response = ServiceClient.sendRequest(HttpMethod.GET,
		                                                     getServer(),
		                                                     getPort(),
		                                                     getPath() + CORE_SERVICES_PATH,
		                                                     queryParameters);

		// ------------------------------- response handling
		if (HttpStatus.OK.value() == response.getStatus()) {
			status = jsonConverter.readValue(response.getData(), SAPRfcTidStatus.class);
		}
		
		// -------------------------- return
		return status;
	}

	/**
	 * SAP RFC service incoming document message handling.
	 *
	 * @param pServiceCode	executed service code
	 * @param pCompanyId	company including activated service identifier
	 * @param pMessage		incoming document message
	 * @return				flag indicating if service incoming message has been handled
	 */
	public boolean handleRfcMessage(String pServiceCode,
	                                Integer pCompanyId,
	                                SAPRfcMessage pMessage) throws JsonProcessingException {
		// ------------------------------- sending the request
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(ACTION_PARAMETER_NAME, HANDLE_RFC_MESSAGE);
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);
		queryParameters.put("serializedRequestBody", jsonConverter.writeValueAsString(pMessage));
		ServiceResponse response = ServiceClient.sendRequest(HttpMethod.FORM_DATA,
		                                                     getServer(),
		                                                     getPort(),
		                                                     getPath() + CORE_SERVICES_PATH,
		                                                     queryParameters);
			
		// -------------------------- return
		return HttpStatus.OK.value() == response.getStatus();
	}

	/**
	 * SAP RFC service server error into database saving.
	 *
	 * @param pServiceCode	executed service code
	 * @param pCompanyId	company including activated service identifier
	 * @param pServerError	server error to be saved
	 * @return				flag indicating if service server error has been saved
	 */
	public boolean saveRfcServerError(String pServiceCode,
	                                  Integer pCompanyId,
	                                  SAPRfcServerError pServerError) throws JsonProcessingException {
		// ------------------------------- sending the request
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(ACTION_PARAMETER_NAME, SAVE_RFC_SERVER_ERROR);
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);
		queryParameters.put("serializedRequestBody", jsonConverter.writeValueAsString(pServerError));
		ServiceResponse response = ServiceClient.sendRequest(HttpMethod.FORM_DATA,
		                                                     getServer(),
		                                                     getPort(),
		                                                     getPath() + CORE_SERVICES_PATH,
		                                                     queryParameters);
			
		// -------------------------- return
		return HttpStatus.OK.value() == response.getStatus();
	}

	/**
	 * Machine reponse duration log saving.
	 *
	 * @param pServiceCode		related service code
	 * @param pCompanyId		company including activated service identifier
	 * @param pEventCode		event related to the machine
	 * @param pResponseDuration	response duration data container
	 * @return
	 */
	public boolean saveMachineResponseDuration(String pServiceCode,
	                                           Integer pCompanyId,
	                                           String pEventCode,
	                                           MachineResponseDuration pResponseDuration) throws JsonProcessingException {
		// ------------------------------- sending the request
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(ACTION_PARAMETER_NAME, SAVE_MACHINE_RESOPONSE_DURATION);
		queryParameters.put(SERVICE_CODE_PARAMETER_NAME, pServiceCode);
		queryParameters.put(COMPANY_ID_PARAMETER_NAME, pCompanyId);
		queryParameters.put(EVENT_CODE_PARAMETER_NAME, pEventCode);
		queryParameters.put("serializedRequestBody", jsonConverter.writeValueAsString(pResponseDuration));
		ServiceResponse response = ServiceClient.sendRequest(HttpMethod.FORM_DATA,
		                                                     getServer(),
		                                                     getPort(),
		                                                     getPath() + CORE_SERVICES_PATH,
		                                                     queryParameters);

	// -------------------------- return
	return HttpStatus.OK.value() == response.getStatus();
	}

	public boolean monitorSystemServicesStatus(StatusContainer serviceContainer) {
		Map<String, Object> queryParameters = new HashMap<>();
		queryParameters.put(ACTION_PARAMETER_NAME, MONITOR_SYSTEM_SERVICES);
		ServiceResponse response = ServiceClient.sendRequest(HttpMethod.POST,
		                                                     getServer(),
		                                                     getPort(),
		                                                     getPath() + CORE_SERVICES_PATH,
		                                                     queryParameters,
		                                                     serviceContainer);

		return HttpStatus.OK.value() == response.getStatus();
	}
}

