package cz.aimtec.dci.coreservices;

import cz.aimtec.dci.coreservices.api.model.*;
import cz.aimtec.dci.microservice.client.HttpMethod;
import cz.aimtec.dci.microservice.client.ServiceClient;
import cz.aimtec.dci.microservice.client.ServiceResponse;
import cz.aimtec.dci.microservice.discovery.DiscoveryService;
import cz.aimtec.util.exception.ApplicationRuntimeException;
import org.apache.http.HttpStatus;

import java.util.*;

/**
 * Consumption core services client.
 */
public class ServiceConsumptionClient extends ServiceClient {

	/**
	 * Basic constructor.
	 *
	 * @param pDiscovery		service discovery provider
	 * @param pServiceName		service name
	 * @param pRequiredVersion	service version
	 */
	public ServiceConsumptionClient(DiscoveryService pDiscovery,
									String pServiceName,
									String pRequiredVersion) {
		super(pDiscovery, pServiceName, pRequiredVersion);
	}

	/**
	 * Basic constructor.
	 *
	 * @param pDiscovery   service discovery provider
	 * @param pServiceName service name
	 */
	public ServiceConsumptionClient(DiscoveryService pDiscovery,
									String pServiceName) {
		super(pDiscovery, pServiceName);
	}

	/**
	 * Basic constructor.
	 *
	 * @param pDiscovery service discovery provider
	 */
	public ServiceConsumptionClient(DiscoveryService pDiscovery) {
		super(pDiscovery);
	}

	/**
	 * Basic constructor.
	 */
	public ServiceConsumptionClient() {
		super();
	}

	/**
	 * All active services registerred at core services service.
	 *
	 * @param pCompanyId company identifier
	 * @return services status container
	 */

	public StatusContainer getActiveServices(Integer pCompanyId) {
		return getActiveServices(pCompanyId, null);
	}

	/**
	 * Service specified by pServiceCode paramater
	 *
	 * @param pCompanyId   company identifier
	 * @param pServiceCode service code
	 * @return services status container
	 */

	public StatusContainer getActiveServices(Integer pCompanyId, String pServiceCode) {

		// ---------------------- send request to core services
		Map<String, Object> parameters = new HashMap<String, Object>(1);
		parameters.put("companyId", pCompanyId);
		parameters.put("serviceCode", pServiceCode);

		return getAdministratorServiceStatusContainer(parameters, "Can not read active services.");
	}

	private StatusContainer getAdministratorServiceStatusContainer(Map<String, Object> parameters, String error) {
		StatusContainer statusContainer = null;
		ServiceResponse response = callMethod(HttpMethod.GET, "administrator/status", parameters);

		// ---------------------- parse response data
		if (response.getStatus() == (HttpStatus.SC_OK)) {
			try {
				statusContainer = getJsonConverter().readValue(response.getData(), StatusContainer.class);

			} catch (Exception pException) {
				throw new ApplicationRuntimeException(error, pException);
			}
		}

		// ---------------------- return
		return statusContainer;
	}

	/**
	 * Service pause.
	 *
	 * @param pCode			service code
	 * @param pCompanyId	company identifier which the service belongs to (null for application services)
	 * @return				flag indicating the action was successful
	 */
	public boolean pauseService(String pCode, Integer pCompanyId) {
		Map<String, Object> parameters = new HashMap<String, Object>(2);
		parameters.put("code", pCode);
		parameters.put("companyId", pCompanyId);
		ServiceResponse response = callMethod(HttpMethod.POST, "service/pause", parameters);

		// ---------------------- return
		return response.getStatus() == HttpStatus.SC_OK;
	}

	/**
	 * Paused service resuming.
	 *
	 * @param pCode			service code
	 * @param pCompanyId	company identifier which the service belongs to (null for application services)
	 * @return				flag indicating the action was successful
	 */
	public boolean resumeService(String pCode, Integer pCompanyId) {
		Map<String, Object> parameters = new HashMap<String, Object>(2);
		parameters.put("code", pCode);
		parameters.put("companyId", pCompanyId);
		ServiceResponse response = callMethod(HttpMethod.POST, "service/resume", parameters);

		// ---------------------- return
		return response.getStatus() == HttpStatus.SC_OK;
	}

	/**
	 * Service restart.
	 *
	 * @param pService		service configuration (contains new interval f.e.)
	 * @return				flag indicating the action was successful
	 */
	public boolean restartService(ServiceConfiguration pService) {
		ServiceResponse response = callMethod(HttpMethod.POST, "service/restart", pService);

		// ---------------------- return
		return response.getStatus() == HttpStatus.SC_OK;
	}

	/**
	 * Service event pause.
	 *
	 * @param pServiceCode	service code
	 * @param pCompanyId	company identifier which the service belongs to (null for application services)
	 * @param pEventCode	event code
	 */
	public void pauseEvent(String pServiceCode, Integer pCompanyId, String pEventCode) {
		Map<String, Object> parameters = new HashMap<String, Object>(2);
		parameters.put("serviceCode", pServiceCode);
		parameters.put("companyId", pCompanyId);
		parameters.put("eventCode", pEventCode);

		// ---------------------- return
		callMethod(HttpMethod.POST, "serviceEvent/pause", parameters);
	}

	/**
	 * Service event restart.
	 *
	 * @param pService		service configuration (contains new event parameters - period f.e.)
	 * @param pEventCode	restarted event code
	 */
	public void restartEvent(ServiceConfiguration pService, String pEventCode) {
		Map<String, Object> parameters = new HashMap<String, Object>(2);
		parameters.put("eventCode", pEventCode);
		// ---------------------- return
		callMethod(HttpMethod.POST, "serviceEvent/restart", pService, parameters);
	}

	/**
	 * Service event command sending.
	 *
	 * @param pCommand			command to be sent to service event
	 * @return					response message container
	 */
	public CommandMessageResponse sendEventCommand(CommandMessage pCommand) {
		// ----------------------------- send request to core services
		ServiceResponse response = callMethod(HttpMethod.POST, "serviceEvent/sendCommand", pCommand);
		CommandMessageResponse commandResponse = parseCommandResponse(response);

		// ----------------------------- return
		return commandResponse;
	}

	/**
	 * Service event command response reading.
	 *
	 * @param pCommand			event command which response should be read
	 * @return					response message container
	 */
	public CommandMessageResponse readEventCommandResponse(CommandMessage pCommand) {
		// ----------------------------- send request to core services
		ServiceResponse response = callMethod(HttpMethod.POST, "serviceEvent/readCommandResponse", pCommand);
		CommandMessageResponse commandResponse = parseCommandResponse(response);

		// ----------------------------- return
		return commandResponse;
	}

	/**
	 * Service event command response checking.
	 *
	 * @param pCommand			event command which response should be checked
	 * @return					flag indicating existing command response
	 */
	public boolean hasEventCommandResponse(CommandMessage pCommand) {
		// ----------------------------- send request to core services
		ServiceResponse response = callMethod(HttpMethod.POST, "serviceEvent/hasCommandResponse", pCommand);

		// ----------------------------- return
		return response.getStatus() == (HttpStatus.SC_OK);
	}

	/**
	 * Method provides all service applicable command names.
	 * It is implemented for TCPIP listener only.
	 *
	 * @param pServiceCode	service code
	 * @param pCompanyId	company identifier which the service belongs to (null for application services)
	 * @return				list of applicable commands names
	 */
	public List<String> getServiceAllowedCommandNames(String pServiceCode,
													  Integer pCompanyId) {
		List<String> commandNames = new ArrayList<String>();

		// ---------------------- send request to core services
		Map<String, Object> parameters = new HashMap<String, Object>(1);
		parameters.put("serviceCode", pServiceCode);
		parameters.put("companyId", pCompanyId);
		ServiceResponse response = callMethod(HttpMethod.GET, "administrator/allowedCommands", parameters);

		// ---------------------- parse response data
		if (response.getStatus() == (HttpStatus.SC_OK)) {
			try {
				String[] commandNamesNative = getJsonConverter().readValue(response.getData(), String[].class);
				commandNames = Arrays.asList(commandNamesNative);

			} catch (Exception exception) {
				throw new ApplicationRuntimeException("Can not read allowed command names.", exception);
			}
		}

		// ---------------------- return
		return commandNames;
	}

	/**
	 * Method provides all service watchdog type names.
	 * It is implemented for TCPIP listener only.
	 *
	 * @return				list of watchdog type names
	 */
	public List<String> getServiceWatchdogTypes() {
		List<String> watchdogTypes = new ArrayList<String>();

		// ---------------------- send request to core services
		ServiceResponse response = callMethod(HttpMethod.GET, "administrator/watchdogTypes", (Object) null);

		// ---------------------- parse response data
		if (response.getStatus() == (HttpStatus.SC_OK)) {
			try {
				String[] watchdogTypesNative = getJsonConverter().readValue(response.getData(), String[].class);
				watchdogTypes = Arrays.asList(watchdogTypesNative);

			} catch (Exception exception) {
				throw new ApplicationRuntimeException("Can not read watchdog types.", exception);
			}
		}

		// ---------------------- return
		return watchdogTypes;
	}

	/**
	 * Service event OPC command sending.
	 *
	 * @param pCommand			command to be sent to service event
	 * @return					response message container
	 */
	public CommandMessageResponse sendOpcCommand(CommandMessage pCommand) {
		// ----------------------------- send request to core services
		ServiceResponse response = callMethod(HttpMethod.POST, "serviceEvent/sendOpcCommand", pCommand);
		CommandMessageResponse commandResponse = parseCommandResponse(response);

		// ----------------------------- return
		return commandResponse;
	}

	/**
	 * Method reads OPC node hierarchy.
	 *
	 * @param pCommand	service code
	 * @return			list of OPC nodes hierarchy
	 */
	public List<OPCNodeHierarchy> readOpcNodeHierarchy(CommandMessage pCommand) {
		// ----------------------------- send request to core services
		ServiceResponse response = callMethod(HttpMethod.POST, "serviceEvent/opcNodeHierarchy", pCommand);
		List<OPCNodeHierarchy> nodeHierarchy = new ArrayList<OPCNodeHierarchy>();

		// ---------------------- parse response data
		if (response.getStatus() == (HttpStatus.SC_OK)) {
			try {
				OPCNodeHierarchy[] nodeHierarchyNative = getJsonConverter().readValue(response.getData(),
																					  OPCNodeHierarchy[].class);
				nodeHierarchy = Arrays.asList(nodeHierarchyNative);

			} catch (Exception exception) {
				throw new ApplicationRuntimeException("Can not read OPC node hierarchy.", exception);
			}

		} else {
			throw new ApplicationRuntimeException("Can not read OPC node hierarchy.");
		}

		// ----------------------------- return
		return nodeHierarchy;
	}

	/**
	 * Service event OPC subscriptions refreshment.
	 *
	 * @param pService	service configuration instance
	 */
	public void refreshOpcSubscriptions(ServiceConfiguration pService) {
		// ----------------------------- send request to core services
		ServiceResponse response = callMethod(HttpMethod.POST, "serviceEvent/refreshOpcSubscriptions", pService);

		// ----------------------------- response handling
		if (response.getStatus() != (HttpStatus.SC_OK)) {
			throw new ApplicationRuntimeException("Can not refresh OPC subscriptions.");
		}
	}

	/**
	 * Get direct children nodes for given rootNode of request in FancyTree node format
	 * that is used for rendering result
	 * 
	 * @param pRequest request POJO
	 * @return Available nodes or error code
	 */
	public BrowseOPCNodesResult browseOpcNodes(BrowseOPCNodesRequest pRequest) {
		// ---------------------- send request to core services
		BrowseOPCNodesResult nodes;
		try {
			ServiceResponse response = callMethod(HttpMethod.POST, "serviceEvent/browseOpcNodes", pRequest);
			nodes = getJsonConverter().readValue(response.getData(), BrowseOPCNodesResult.class);
		} catch (Exception ex) {
			throw new ApplicationRuntimeException("Can not browse OPC nodes", ex);
		}
		
		return nodes;
	}

	private CommandMessageResponse parseCommandResponse(ServiceResponse pResponse) {
		CommandMessageResponse commandResponse = new CommandMessageResponse();

		try {
			if (pResponse.getStatus() == (HttpStatus.SC_OK)) {
				commandResponse = getJsonConverter().readValue(pResponse.getData(), CommandMessageResponse.class);
			}

		} catch (Exception exception) {
			throw new ApplicationRuntimeException("Can not read service event command response.", exception);
		}

		// -------------------------- return
		return commandResponse;
	}
}
