package cz.aimtec.dci.microservices;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import cz.aimtec.dci.coregate.coreservices.ServiceConsumptionClient;
import cz.aimtec.dci.coreservices.api.model.ServiceConfiguration;
import cz.aimtec.dci.coreservices.model.IncomingMessage;
import cz.aimtec.dci.microservice.discovery.DiscoveryService;
import cz.aimtec.dci.microservice.eventbus.EventBusClient;
import cz.aimtec.util.exception.ApplicationRuntimeException;
import org.springframework.beans.factory.annotation.Autowired;

import java.util.ArrayList;
import java.util.List;

/**
 * Core gate service client.
 *
 * @author hors
 * @version 11. 10. 2017/hors - created<br>
 */
public class CoreGate extends ServiceConsumptionClient {

	@Autowired
	private ObjectMapper jsonConverter;

	public CoreGate() {
		super();
	}

	/**
	 * Basic constructor.
	 *
	 * @param pDiscovery service discovery provider
	 */
	public CoreGate(DiscoveryService pDiscovery,
					String pServiceName) {
		super(pDiscovery, null, pServiceName);
	}

	/**
	 *
	 * @param pDiscovery 		service discovery provider
	 * @param eventBusClient	EventBus client
	 * @param pServiceName		service name
	 */
	public CoreGate(DiscoveryService pDiscovery,
					EventBusClient eventBusClient,
					String pServiceName) {
		super(pDiscovery, eventBusClient, pServiceName);
	}

	/**
	 *
	 * @param pDiscovery		service discovery provider
	 * @param eventBusClient	EventBus client
	 * @param pServiceName		service name
	 * @param pServiceVersion	(deprecated) service version
	 * //TODO pServiceVersion should be read from pom, but its not working here
	 */
	public CoreGate(DiscoveryService pDiscovery,
					EventBusClient eventBusClient,
					String pServiceName,
					String pServiceVersion) {
		super(pDiscovery, eventBusClient, pServiceName, pServiceVersion);
	}

	/**
	 * Method provides all core services to be activated.
	 *
	 * @return	services for activation or null (if it is not possible to get the list)
	 */
	public List<ServiceConfiguration> retrieveServicesConfiguration() throws JsonProcessingException {
		List<ServiceConfiguration> servicesConfiguration = null;

		List<cz.aimtec.dci.coregate.api.model.coreservices.ServiceConfiguration> nativeServiceConfigurations =
				getServicesConfiguration();

		// ---------------------- conversion to core services configurations
		if (nativeServiceConfigurations != null) {
			servicesConfiguration = new ArrayList<>();

			for (cz.aimtec.dci.coregate.api.model.coreservices.ServiceConfiguration nativeServiceConfiguration : nativeServiceConfigurations) {
				servicesConfiguration.add(createServiceConfiguration(nativeServiceConfiguration));
			}
		}

		// ---------------------- return
		return servicesConfiguration;
	}

	/**
	 * Method handles incoming listener message.
	 *
	 * @param pMessage		incoming message to be handled
	 */
	public void handleMessage(IncomingMessage pMessage) {
		// ------------------- message conversion to core gate one
		cz.aimtec.dci.coregate.api.model.coreservices.IncomingMessage message = createIncomingMessage(pMessage);

		// ---------------------- send request to core gate
		super.handleMessage(message);
	}

	/**
	 * Messages archiving.
	 * It moves active incoming event messages into archive table.
	 *
	 * @param pMessages	list of messages to be archived
	 */
	public void sendMessagesToArchive(List<IncomingMessage> pMessages) {
		// ------------------- message conversion to core gate one
		List<cz.aimtec.dci.coregate.api.model.coreservices.IncomingMessage> messages = new ArrayList<>();

		for (IncomingMessage nativeMessage : pMessages) {
			messages.add(createIncomingMessage(nativeMessage));
		}

		// ---------------------- send request to core gate
		super.archiveMessages(messages);
	}

	private ServiceConfiguration createServiceConfiguration(cz.aimtec.dci.coregate.api.model.coreservices.ServiceConfiguration pConfiguration) {
		ServiceConfiguration configuration = null;

		try {
			if (pConfiguration != null) {
				configuration = getJsonConverter().readValue(jsonConverter.writeValueAsString(pConfiguration),
															 ServiceConfiguration.class);
			}
		} catch (Exception exception) {
			throw new ApplicationRuntimeException("Can not create service configuration.", exception);
		}

		// ------------------------- return
		return configuration;
	}

	private cz.aimtec.dci.coregate.api.model.coreservices.IncomingMessage createIncomingMessage(IncomingMessage pMessage) {
		cz.aimtec.dci.coregate.api.model.coreservices.IncomingMessage message = null;

		try {
			message = getJsonConverter().readValue(jsonConverter.writeValueAsString(pMessage),
												   cz.aimtec.dci.coregate.api.model.coreservices.IncomingMessage.class);
		} catch (Exception exception) {
			throw new ApplicationRuntimeException("Can not create incoming message.", exception);
		}

		return message;
	}

	public EventBusClient getEventBus() {
		return this.eventBusClient;
	}
}

