package cz.aimtec.dci.coreservices.listener.tcpip.opc;

import com.prosysopc.ua.ServiceException;
import com.prosysopc.ua.StatusException;
import com.prosysopc.ua.client.MonitoredDataItem;
import com.prosysopc.ua.client.MonitoredDataItemListener;
import com.prosysopc.ua.client.UaClient;
import com.prosysopc.ua.stack.builtintypes.DataValue;
import com.prosysopc.ua.stack.builtintypes.NodeId;
import com.prosysopc.ua.stack.builtintypes.Variant;
import com.prosysopc.ua.stack.core.Attributes;
import cz.aimtec.dci.coreservices.implementation.TCPIPListenerEvent;
import cz.aimtec.dci.coreservices.listener.MessageDecoder;
import cz.aimtec.dci.coreservices.listener.MessageListenerClient;
import cz.aimtec.dci.coreservices.listener.MessageType;
import cz.aimtec.dci.coreservices.listener.tcpip.textReader.MultiValueTextReaderParsingAttributes;
import cz.aimtec.dci.coreservices.model.IncomingMessage;
import cz.aimtec.util.Convert;
import cz.aimtec.util.datatype.DateTime;
import cz.aimtec.util.exception.ApplicationRuntimeException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.UUID;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * 
 * Default transaction execution from OPC node subscription
 * @author asel
 * 
 */
public class DefaultTransMonitoredDataItemListener implements MonitoredDataItemListener {
	
	public static final String AUTHOR_SUBSCRIPTION = "subscription";
	
	private UaClient client;
	private TCPIPListenerEvent event;
	private boolean lastValueSet = false;
	private OPCNode opcNode = null;
	private DateTime lastRunTime = null;
	private final UUID uid = UUID.randomUUID();
	private boolean subscriptionChecking = false;
	private int publishingInterval = 0;
		
	/* If there is any connection problem this thread is invalidated
	 * When reconnected it cannot be known if subscriptions were discarted or not (both situations are possible and both are valid)
	 */ 
	private boolean isValid = true;
	
	private static Logger log = LoggerFactory.getLogger(DefaultTransMonitoredDataItemListener.class);


	/**
	 * Constructor.
	 *
	 * @param client               native OPC client
	 * @param event                listener event
	 * @param opcNode              OPC node. It contains what transaction definition should be run, machine and attribute.
	 * @param subscriptionChecking
	 * @param publishingInterval
	 */
	public DefaultTransMonitoredDataItemListener(UaClient client, TCPIPListenerEvent event, OPCNode opcNode, boolean subscriptionChecking, int publishingInterval) {
		this.client = client;
		this.event = event;
		this.opcNode = opcNode;
		this.isValid = true;
		this.subscriptionChecking = subscriptionChecking;
		this.publishingInterval = publishingInterval;

		MessageListenerClient.registerListenerCodeForLogging(event.getClient());
	}

	
	public TCPIPListenerEvent getEvent() {
		return event;
	}

	public DateTime getLastRunTime() {
		return this.lastRunTime;
	}
	
	public OPCNode getOpcNode() {
		return this.opcNode;
	}
	
	private synchronized boolean isLastValueSet() {
		return this.lastValueSet;
	}
	
	private synchronized void setLastValueSet(boolean lastValueSet) {
		this.lastValueSet = lastValueSet;
	}

	public boolean isValid() {
		return this.isValid;
	}

	public int getPublishingInterval() {
		return publishingInterval;
	}
	
	
	@Override
	public void onDataChange(MonitoredDataItem sender, DataValue prevValue, DataValue value) {
		if (!this.isValid) {
			return;
		}

		Variant valueVariant = value.getValue();
		String valueString = Convert.toString(valueVariant.getValue());
		
		// -------------------- parse nodeId
		NodeId nodeId = NodeId.parseNodeId(opcNode.getOpcNodeId());
		OPCDataType attributeDataType;
		
		if (isExcluded(valueString) || valueVariant.isEmpty()) {
			log.info("Excluding subscription nodeId={}, value={}", nodeId, valueString);
			return;
		}
		
		String dataTypeIdentifier;
		
		try {
			dataTypeIdentifier = client.readAttribute(nodeId, Attributes.DataType).getValue().getValue().toString();
			attributeDataType = OPCDataType.getDataTypeByIdCode(dataTypeIdentifier);
		} catch (ServiceException | StatusException exception) {
			throw new ApplicationRuntimeException("Can not get OPC attribute value type.", exception);
		}
		
		// --------------------- value change handling
		boolean nodeHasTransaction = Convert.isNotBlank(opcNode.getTransactionDefinitionName());
		String lastDBValue = opcNode.getLastSubscriptionValue();
		String dataVector = attributeDataType.toString(value.getValue());
		
		if (log.isDebugEnabled()) {
			log.debug("onDataChange: Start, Subscription UID={}", uid);
		}
		
		if (!this.isLastValueSet() && nodeHasTransaction) {
			this.setLastValueSet(true);
			if (subscriptionChecking && value.getValue() != null && !dataVector.equals(lastDBValue)) {
				handleIncomingMessage(dataVector, dataTypeIdentifier);
			}
			
			log.debug("Subscription onDataChange: Initialized. NodeId={}", opcNode.getOpcNodeId());
			
		} else {
			log.debug("Subscription onDataChange: Value changed - Starting. NodeId={}", opcNode.getOpcNodeId());
			
			if (nodeHasTransaction) {
				handleIncomingMessage(dataVector, dataTypeIdentifier);
			} else {
				log.warn("Subscription onDataChange: Warning! Transaction run skipped, there is no transaction set for this node.");
			}
			
			log.debug("Subscription onDataChange: Finished. NodeId={}", opcNode.getOpcNodeId());
		}
		
		this.lastRunTime = DateTime.now();
	}
	
	private void handleIncomingMessage(String pDataVector,
	                                   String pDataTypeIdentifier) {
		// ------------------------- last subscription value saving
		opcNode.setLastSubscriptionValue(pDataVector);
		
		// ------------------------- opc event logging
		OPCClient opcClient = (OPCClient) getEvent().getClient();
		opcClient.logOpcEvent(opcNode.getOpcNodeId(), opcNode.getMachineCode(), pDataVector, AUTHOR_SUBSCRIPTION);
		
		// ------------------------- message handling
		MessageDecoder decoder = getEffectiveMessageDecoder();
		MessageType messageType = Convert.isNotBlank(decoder) ? decoder.getFirstMessageType(false) : null;
		
		if (messageType != null) {
			IncomingMessage message = new IncomingMessage(getEvent(),
			                                              getEvent().getListener(),
			                                              messageType,
			                                              Convert.getStringAsBytes(pDataVector),
			                                              pDataVector,
			                                              null,
			                                              getOpcNode().getOpcNodeId(),
			                                              pDataTypeIdentifier,
														  getOpcNode().getMachineCode());
			message.setTransactionDefinitionName(opcNode.getTransactionDefinitionName());
			messageType.getMessageHandling().handleIncomingMessage(message, getEvent());
		}
	}
	
	/**
	 * Make this listener invalid.
	 */
	public void invalidate() {
		log.debug("Invalidating OPC subscription. NodeId={}.", opcNode.getOpcNodeId());
		this.isValid = false;
	}
	
	private boolean isExcluded(String dataVector) {
		MessageDecoder messageDecoder = getEffectiveMessageDecoder();
		
		if (Convert.isNull(messageDecoder)) {
			return false;
		}
		
		MessageType messageType = messageDecoder.getFirstMessageType(false);
		if (Convert.isNull(messageType) || Convert.isBlank(messageType.getParsingAttributes())) {
			return false;
		}
		
		MultiValueTextReaderParsingAttributes parsingAttributes = MultiValueTextReaderParsingAttributes.parseDefinition(messageType.getParsingAttributes());
		String messageExclude = parsingAttributes.getFields().get(0).getMessageExclude();
		
		if (Convert.isNotBlank(messageExclude)) {
			Matcher reqularExpressionMatcher = Pattern.compile(messageExclude).matcher(dataVector);

			if (reqularExpressionMatcher.find()) {
				return true;
			}
		}
		
		return false;
	}
	
	private MessageDecoder getEffectiveMessageDecoder() {
		return Convert.isNotBlank(opcNode.getMessageDecoder()) ? opcNode.getMessageDecoder() : getEvent().getListener().getMessageDecoder();
	}
}
