package cz.aimtec.dci.coreservices.listener.tcpip.opc;

import com.prosysopc.ua.*;
import com.prosysopc.ua.client.AddressSpaceException;
import com.prosysopc.ua.client.MonitoredDataItem;
import com.prosysopc.ua.client.Subscription;
import com.prosysopc.ua.client.UaClient;
import com.prosysopc.ua.nodes.UaNode;
import com.prosysopc.ua.stack.builtintypes.*;
import com.prosysopc.ua.stack.cert.DefaultCertificateValidator;
import com.prosysopc.ua.stack.cert.PkiDirectoryCertificateStore;
import com.prosysopc.ua.stack.common.ServiceResultException;
import com.prosysopc.ua.stack.core.*;
import com.prosysopc.ua.stack.transport.AsyncResult;
import com.prosysopc.ua.stack.transport.security.SecurityMode;
import cz.aimtec.dci.coreservices.implementation.TCPIPListenerEvent;
import cz.aimtec.dci.coreservices.listener.MessageListener;
import cz.aimtec.dci.coreservices.listener.MessageListenerClient;
import cz.aimtec.dci.coreservices.listener.WatchdogType;
import cz.aimtec.dci.coreservices.listener.tcpip.TCPIPClient;
import cz.aimtec.dci.coreservices.service.MultiEventService;
import cz.aimtec.util.Convert;
import cz.aimtec.util.HashList;
import cz.aimtec.util.datatype.DateTime;
import cz.aimtec.util.exception.ApplicationException;
import org.apache.mina.core.service.IoHandlerAdapter;
import org.apache.mina.filter.codec.ProtocolCodecFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.env.Environment;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.text.SimpleDateFormat;
import java.util.*;

/**
 * OPC Client interface. Simplifies UaClient interface for DCI purposes.
 * @author asel
 * 
 */
public class OPCClient extends TCPIPClient {
	
	private static final String DATE_FORMAT_MASK = "HH:mm:ss";
	private static final int	OPC_CLIENT_PUBLISHING_INTERVAL_DEFAULT = 1000;
	private static final int	OPC_CLIENT_SESSION_TIMEOUT_DEFAULT = 3600000;
	private static final int	OPC_CLIENT_LIFE_TIME_COUNT_AS_TIME = 60000;
	private static final int	OPC_CLIENT_KEEP_ALIVE_COUNT_AS_TIME = 5000;
	private static final int	OPC_CLIENT_TIMEOUT_DEFAULT = 5000;

	private static final String TIME = ", time: ";


	/**
	 * Set up client before connection is established.
	 * Set OPC server address, description and other mandatory fields.
	 * Connection settings is read from settings given in constructor.
	 * @throws SessionActivationException
	 *
	 */
	private void setup() throws URISyntaxException, SecureIdentityException, IOException, SessionActivationException {
		log.debug("OPCClient.setup ... starting");

		MessageListener listener = getEvent().getListener();

		// Server address
		log.debug("OPCClient.setup ... getting server address");
		String serverAddress = OPC_TPC_PROTOCOL + listener.getIpAddress() + ":" + listener.getPort()
				+ (listener.getAddressPath() == null ? Convert.EMPTY_STRING : listener.getAddressPath());
		log.debug("OPCClient.setup ... server address: {}", serverAddress);


		client = new UaClient(serverAddress);

		StringBuilder applicationName = new StringBuilder();
		applicationName.append(listener.getCode());
		applicationName.append("_DCIx");

		// Application description
		log.debug("OPCClient.setup ... creating client identity");
		ApplicationDescription appDescription = new ApplicationDescription();
		Environment environment = getEvent().getService().getEnvironment();
		String applicationURI;
		try {
			applicationURI = environment.getProperty("applicationDomainName", "localhost");
		} catch (Exception e) {
			applicationURI = "";
		}
		applicationURI = applicationURI.replaceAll("\"", "");
		if (Convert.isBlank(applicationURI)  || "dci-".equalsIgnoreCase(applicationURI)) {
			applicationURI = "localhost";
		}
		appDescription.setApplicationName(new LocalizedText(applicationName.toString(), Locale.ENGLISH));
		appDescription.setApplicationUri("urn:" + applicationURI + ":UA:" + listener.getCode());
		appDescription.setProductUri("urn:aimtec.cz:UA:" + listener.getCode());
		appDescription.setApplicationType(ApplicationType.Client);
		log.debug("OPCClient.setup ... description: " + listener.getDescription() + ", code: " + listener.getCode());

		// Server certificate validation listener
		log.debug("OPCClient.setup ... setting up server listener");
		final PkiDirectoryCertificateStore certStore = new PkiDirectoryCertificateStore();
		final DefaultCertificateValidator validator = new DefaultCertificateValidator(certStore);

		client.setCertificateValidator(validator);
		validator.setValidationListener(new OPCServerCertificateValidationListener());
		log.debug("OPCClient.setup ... server listener set");

		File certificate = null;
		log.debug("OPCClient.setup ... creating client certificate");
		if (Convert.isNotBlank(listener.getExternalCertificatePath())) {
			log.debug("OPCClient.setup ... using external certificate");
			certificate = new File(listener.getExternalCertificatePath());
		} else {
			log.debug("OPCClient.setup ... using private certificate");
			certificate = new File(certStore.getBaseDir(), "private");
		}

		log.debug("OPCClient.setup ... creating application identity (client identity + certificate)");
		final ApplicationIdentity identity = ApplicationIdentity.loadOrCreateCertificate(
				appDescription,
				"DCIx",
				"private314password",								/* Private Key Password */
				certificate,										/* Key File Path */
				true);												/* Enable renewing the certificate */

		if (Convert.isNotBlank(listener.getUserName())) {
			UserIdentity userIdentity = new UserIdentity(listener.getUserName(), listener.getPassword());
			client.setUserIdentity(userIdentity);
		}

		log.debug("OPCClient.setup ... reading watchdog bypass settings");
		int listenerCheckTimeout = getEvent().getListener().getConnectionCheckTimeout();
		int listenerConnectionTimeout = getEvent().getListener().getConnectionTimeout();
		boolean endPointsValidation = getEvent().getListener().isEndPointsValidation();

		int timeoutValue = convertCfgValue(listenerCheckTimeout,  OPC_CLIENT_TIMEOUT_DEFAULT);
		log.debug("OPCClient.setup ... bypass settings: " + timeoutValue);

		log.debug("OPCClient.setup ... setting up client");
		int sessionTimeoutValue = convertCfgValue(getEvent().getListener().getSessionTimeout(), OPC_CLIENT_SESSION_TIMEOUT_DEFAULT);
		client.setApplicationIdentity(identity);

		log.debug("OPCClient.setup ... security mode: " + listener.getSecurityMode());
		client.setSecurityMode(OPCSecurityMode.parseSecurityMode(listener.getSecurityMode()));

		log.debug("OPCClient.setup ... timeout: " + listenerConnectionTimeout);
		client.setTimeout(listenerConnectionTimeout); // communication timeout for service operation, default 120 seconds
		client.setConnectTimeout(listenerConnectionTimeout); // default 60 seconds

		log.debug("OPCClient.setup ... validate endpoints: " + endPointsValidation);
		client.setValidateDiscoveredEndpoints(endPointsValidation);

		client.setStatusCheckTimeout(timeoutValue); // how long Status message can be delayed before server status is changed
		client.setStatusCheckInterval(1000); // Status message interval
		client.setSessionTimeout(sessionTimeoutValue);
		// When autoReconnect=false -> beware of client.isConnected() method that is not working and returns true even when client is disconnected!!
		client.setAutoReconnect(true);

		client.addServerStatusListener(new OPCServerStatusListener(this));
		log.debug("OPCClient.setup ... set up ready");
	}

	private UaClient client = null;
	public static final String OPC_TPC_PROTOCOL = "opc.tcp://";
	private static final Logger log = LoggerFactory.getLogger(OPCClient.class);


	/**
	 * Basic constructor.
	 *
	 * @param pEvent TCPIP listener event
	 */
	public OPCClient(TCPIPListenerEvent pEvent) {
		super(pEvent);
	}

	/**
	 * Fake constructor for jUnit
	 */
	public OPCClient() {
		super();
	}


	public UaClient getClient() {
		return client;
	}


	@Override
	public ProtocolCodecFactory createCodecFactory(TCPIPListenerEvent pEvent) {
		return null;    // ?
	}

	@Override
	public IoHandlerAdapter createProtocolHandler() {
		return null;    // ?
	}

	private OPCSubscriptionManager getOPCSubscriptionManager() {
		return ((TCPIPListenerEvent) getEvent()).getOPCSubscriptionManager();
	}

	/**
	 * Starts client. Establishes a connection to OPC server
	 *
	 * @throws ApplicationException
	 */
	@Override
	public void start() throws ApplicationException {
		MessageListenerClient.registerListenerCodeForLogging(this);

		log.debug("OPCClient.start ... starting");

		if (!this.isConnected()) {
			try {
				log.debug("OPCClient.start ... setting up");
				this.setup();

				log.debug("OPCClient.start ... connecting");
				client.connect();

				log.debug("OPCClient.start " + getEvent().getCode() + ":\n" + client.getEndpointConfiguration().toString() + "\n\n");

			} catch (ServiceException | IOException | URISyntaxException | SecureIdentityException e) {
				log.error("Cannot connect to OPC client. " + e.getMessage()
						+ (e.getCause() != null ? " - " + e.getCause().getMessage() : Convert.EMPTY_STRING));
				log.debug("Cannot connect to OPC client. ", e);
				throw new ApplicationException("Cannot start OPC client.", e);
			}
		}
		log.debug("OPCClient.start ... started");
	}

	/**
	 * Stops client. Disconnects client from OPC server.
	 */
	@Override
	public void stop() {
		log.debug("OPCClient.stop ... stopping (unsubscribe all)");
		try {
			this.unsubscribeAll();
		} catch (ServiceException e) {
			log.error("Error during removing subscriptions", e);
		}

		log.debug("OPCClient.stop ... force disconnect");
		this.forceDisconnect();

		log.debug("OPCClient.stop ... stopped");
	}

	private void forceDisconnect() {
		try {
			if (client != null) {
				log.debug("OPCClient.forceDisconnect ... disconnect");
				client.disconnect();
			}
		} catch (Exception e) {
			log.debug("OPC disconnect failed. It is probably already disconnected (this is not an error, just a debug information).", e);
		}
	}

	/**
	 * For OPC, isConnected() is enough. most of the time these types of packets are disabled by firewall anyway,
	 * so it returns false even when isConnected() returns true.
	 */
	@Override
	public WatchdogType findWatchdogType(MessageListener pListener) {
		WatchdogType watchdogType = pListener.getWatchdogType();
		return WatchdogType.OPCCheckBasedOnNodeReading.equals(watchdogType) ? WatchdogType.OPCCheckBasedOnNodeReading : WatchdogType.OPCCheck;
	}

	/**
	 * Return value if value is valid positive integer. Otherwise return defaultValue.
	 */
	private int convertCfgValue(int value, int defaultValue) {
		return value > 0 ? value : defaultValue;
	}

	/**
	 * Returns if client is connected to server
	 * @return
	 */
	@Override
	public boolean isConnected() {
		log.debug("OPCClient.isConnected ... starting");
		
		// Test if client is initialized

		if (client == null) {
			log.debug("OPCClient.isConnected ... client is null!");
			return false;
		}
		
		// Try standard connection test (soft test - not reliable)
		if (!client.isConnected()) {
			log.debug("OPCClient.isConnected ... soft test failed");
			return false;
		}
		
		return ServerState.Running.equals(client.getServerState()) || ServerState.Test.equals(client.getServerState());
	}

	public boolean isConnectedBasedOnNodeReading() {
		MessageListener listener = getEvent().getListener();
		
		// Try to read node form OPC server (hard test)
		log.debug("OPCClient.isConnected ... reading test nodeId");
		String nodeId = (listener != null) ? listener.getWatchdogSettings() : null;
		log.debug("OPCClient.isConnected ... test nodeId={}.", nodeId);
		
		if (Convert.isBlank(nodeId)) {
			log.debug("OPCClient.isConnected ... nodeId is blank");
			return true;
		}
		AsyncResult<ReadResponse> r = null;
		
		log.debug("OPCClient.isConnected ... read test (async)");
		try {
			r = client.readAsync(
					Double.valueOf(0),
					TimestampsToReturn.Both,
					new ReadValueId(NodeId.parseNodeId(nodeId), Attributes.NodeId, null, null)
					);
		} catch (Exception e) {
			log.debug("OPCClient.isConnected ... read test failed (exception)", e);
			return false;
		}
		
		if (r == null) {
			log.debug("OPCClient.isConnected ... async result is null");
			return false;
		}

		try {
			// Read OPC watchdog timeout
			log.debug("OPCClient.isConnected ... reading timeout settings");
			Integer timeout = getEvent().getListener().getAccessibilityCheckTimeout();
			log.debug("OPCClient.isConnected ... timeout settings: " + timeout);
						
			log.debug("OPCClient.isConnected ... waiting for response");
			ServiceResponse response = r.waitForResult(timeout.intValue(), java.util.concurrent.TimeUnit.MILLISECONDS);
			log.debug("OPCClient.isConnected ... response recieved");
			
			if (response != null
					&& response.getResponseHeader() != null
					&& response.getResponseHeader().getServiceResult() != null
					&& response.getResponseHeader().getServiceResult().isGood()) {
				log.debug("OPCClient.isConnected ... response is 'Good'");
				return true;
			}

			String goodIndicator = "1";
			String badIndicator = "0";
			log.debug("OPCClient.isConnected ... response is 'Bad', internal flags: "
								+ (response != null ? goodIndicator : badIndicator)
								+ (response != null && response.getResponseHeader() != null ? goodIndicator : badIndicator)
								+ (response != null && response.getResponseHeader().getServiceResult() != null ? goodIndicator : badIndicator)
								+ (response != null && response.getResponseHeader().getServiceResult().isGood() ? goodIndicator : badIndicator));
			
		} catch (ServiceResultException e) {
			// timeout occurred
			log.debug("OPC " + listener.getCode() + ".isConnected() - timeout has occured.", e);
			return false;
		}
		
		log.debug("OPCClient.isConnected ... hard read failed = not connected");
		return false;
	}

	/**
	 * @param opcNode
	 * @throws ServiceException
	 * @throws StatusException
	 */
	public void subscribe(OPCNode opcNode) throws ServiceException, StatusException {
		MessageListener listener = getEvent().getListener();
		SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT_MASK);

		log.debug("OPC subscribe: Start, NodeId: " + opcNode.getOpcNodeId());
		NodeId nodeId = null;
		try {
			nodeId = NodeId.parseNodeId(opcNode.getOpcNodeId());
		} catch (Exception e) {
			log.error("Cannot subscribe node: " + Convert.toString(opcNode.getOpcNodeId()), e);
		}

		Calendar timeBefore = Calendar.getInstance();
		log.debug("<before subscribe> nodeId: " + opcNode.getOpcNodeId() + TIME + sdf.format(timeBefore.getTime()) + "."
				+ timeBefore.get(Calendar.MILLISECOND));

		Subscription subscription = new Subscription();

		int publishingInterval = convertCfgValue(getEvent().getListener().getPublishInterval(), OPC_CLIENT_PUBLISHING_INTERVAL_DEFAULT);

		// If OPC node value is available, override default
		if (Convert.isNotBlank(opcNode.getPublishingInterval()) && opcNode.getPublishingInterval().intValue() > 0) {
			publishingInterval = opcNode.getPublishingInterval().intValue();
		}

		subscription.setPublishingEnabled(true);
		subscription.setPublishingInterval(publishingInterval);
		subscription.setLifetimeCount(OPC_CLIENT_LIFE_TIME_COUNT_AS_TIME / publishingInterval);
		subscription.setMaxKeepAliveCount((long) OPC_CLIENT_KEEP_ALIVE_COUNT_AS_TIME / publishingInterval);

		OPCSubscriptionAliveListener aliveListener = new OPCSubscriptionAliveListener(this);
		subscription.addAliveListener(aliveListener);

		OPCSubscriptionNotificationListener subscriptionListener = new OPCSubscriptionNotificationListener(this);
		subscription.addNotificationListener(subscriptionListener);

		MonitoredDataItem item = new MonitoredDataItem(nodeId, UnsignedInteger.valueOf(opcNode.getOpcNodeAttributeId().intValue()),
				MonitoringMode.Reporting);

		boolean subscriptionChecking = (listener == null) ? false : listener.getSubscriptionChecking();
		DefaultTransMonitoredDataItemListener dataChangeListener = new DefaultTransMonitoredDataItemListener(client,
				getEvent(),
				opcNode,
				subscriptionChecking,
				publishingInterval);

		item.setDataChangeListener(dataChangeListener);
		subscription.addItem(item);
		client.addSubscription(subscription);
		getOPCSubscriptionManager().setSubscriptionInfo(this, opcNode, subscription, dataChangeListener);

		Calendar timeAfter = Calendar.getInstance();
		log.debug("<after subscribe> nodeId: " + opcNode.getOpcNodeId() + TIME + sdf.format(timeAfter.getTime()) + "."
				+ timeAfter.get(Calendar.MILLISECOND));
		log.debug("OPC subscribe: End, NodeId: " + opcNode.getOpcNodeId());
	}

	/**
	 * Reads specific nodeId from OPC server. Reads value attribute.
	 * @param nodeIdString
	 * @return
	 * @throws ServiceException
	 * @throws StatusException
	 */
	public Variant read(String nodeIdString, UnsignedInteger attributeID) throws IllegalArgumentException, ServiceException, StatusException {
		SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT_MASK);
		
		NodeId nodeId = NodeId.parseNodeId(nodeIdString);
		
		Calendar timeBefore = Calendar.getInstance();
		log.debug("<before read> nodeId: " + nodeIdString + TIME + sdf.format(timeBefore.getTime()) + "." 
				+ timeBefore.get(Calendar.MILLISECOND));
		
		DataValue value = client.readAttribute(nodeId, attributeID);
		
		log.debug("<after read> nodeId: " + nodeIdString + "; "
				+ "statusCodeName=" + value.getStatusCode().getName() + "; "
				+ "statusCodeDescription=" + value.getStatusCode().getDescription());
		
		Calendar timeAfter = Calendar.getInstance();
		log.debug("<after read> nodeId: " + nodeIdString + TIME + sdf.format(timeAfter.getTime()) + "." 
				+ timeAfter.get(Calendar.MILLISECOND));
		return value.getValue();
	}
	
	/**
	 * Reads the <i>value</i> attribute of a single node.
	 * 
	 * @param node node ID
	 * @return
	 * @throws ServiceException
	 */
	public DataValue readValue(String node) throws ServiceException {
	    
	    return readValues(Arrays.asList(node)).get(0);
	}

	/**
	 * Reads the value attribute for a list of node ids with a single read
	 * @param nodes list of node ids
	 * @return
	 * @throws ServiceException
	 */
	public List<DataValue> readValues(List<String> nodes) throws ServiceException {
		SimpleDateFormat sdf = new SimpleDateFormat("HH:mm:ss");
		NodeId[] nodesArray = new NodeId[nodes.size()];
		
		for (int i = 0; i < nodes.size(); i++) {
			nodesArray[i] = NodeId.parseNodeId(nodes.get(i));
		}
		
		Calendar timeBefore = Calendar.getInstance();
		log.debug("<before readValues> number of nodes: " + nodes.size() + TIME + sdf.format(timeBefore.getTime()) + "." 
				+ timeBefore.get(Calendar.MILLISECOND));
		
		DataValue values[] = client.readValues(nodesArray, (TimestampsToReturn) null);
		
		Calendar timeAfter = Calendar.getInstance();
		log.debug("<after readValues> number of nodes: " + nodes.size() + TIME + sdf.format(timeAfter.getTime()) + "." 
				+ timeAfter.get(Calendar.MILLISECOND));
		
		return Arrays.asList(values);
	}
	
	/**
	 * Check if given Node ID is valid and is set on OPC server
	 * @param nodeIdString
	 * @return
	 */
	public boolean isNodeValid(String nodeIdString) {
		SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT_MASK);
		
		// Try to parse nodeIdString
		NodeId nodeId = null;
		try {
			nodeId = NodeId.parseNodeId(nodeIdString);
		} catch (Exception e) {
			// If exception occurred nodeIdString is not valid Node ID.
			log.debug("Cannot parse nodeId for validity check.", e);
			return false;
		}
		
		// Try to read node from OPC server
		UaNode node = null;
		try {
			Calendar timeBefore = Calendar.getInstance();
			log.debug("<before node valid> nodeId: " + nodeIdString + TIME + sdf.format(timeBefore.getTime()) + "." 
					+ timeBefore.get(Calendar.MILLISECOND));
			
			node = client.getAddressSpace().getNode(nodeId);
			
			Calendar timeAfter = Calendar.getInstance();
			log.debug("<after node valid> nodeId: " + nodeIdString + TIME + sdf.format(timeAfter.getTime()) + "." 
					+ timeAfter.get(Calendar.MILLISECOND));
		} catch (ServiceException | AddressSpaceException e) {
			log.error("Node cannot be checked or is not valid.", e);
			log.error(Convert.toString(node) + " node cannot be found in address space");
			log.debug("Cannot read nodeId for validity check.", e);
			return false;
		}
		
		// If node is not null it means it was read from OPC server and the node is valid
		return node != null;
	}
	
	/**
	 * Writes specific nodeId to OPC server. Writes value attribute.
	 * @param nodeIdString
	 * @param value
	 * @throws ServiceException
	 * @throws StatusException
	 */
	public void write(String nodeIdString, Object value, UnsignedInteger attributeID) throws ServiceException, StatusException {
		SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT_MASK);

		NodeId nodeId = NodeId.parseNodeId(nodeIdString);
		
		String dataType = client.readAttribute(nodeId, Attributes.DataType).getValue().getValue().toString();

		log.debug("Node: " + nodeId.getValue() + ". Writing node type: " + dataType);

		Calendar timeBefore = Calendar.getInstance();
		log.debug("<before write> nodeId: " + nodeIdString + TIME + sdf.format(timeBefore.getTime()) + "." 
				+ timeBefore.get(Calendar.MILLISECOND));
		
		client.writeAttribute(nodeId, attributeID, getDataTypeByIdCode(dataType).convert(value));
		
		Calendar timeAfter = Calendar.getInstance();
		log.debug("<after write> nodeId: " + nodeIdString + TIME + sdf.format(timeAfter.getTime()) + "." 
				+ timeAfter.get(Calendar.MILLISECOND));
	}

	/**
	 *
	 * @param nodes
	 * @param values
	 * @throws ServiceException
	 * @throws StatusException
	 */
	public void writeValues(List<String> nodes, List<Object> values) throws ServiceException, StatusException {
		SimpleDateFormat sdf = new SimpleDateFormat("HH:mm:ss");

		NodeId[] nodesArray = new NodeId[nodes.size()];
		
		for (int i = 0; i < nodes.size(); i++) {
			nodesArray[i] = NodeId.parseNodeId(nodes.get(i));
		}
		
		Object[] valuesArray = new Object[values.size()];
		
		for (int i = 0; i < values.size(); i++) {
			valuesArray[i] = values.get(i);
		}
		
		Calendar timeBefore = Calendar.getInstance();
		log.debug("<before writeValues> time: " + sdf.format(timeBefore.getTime()) + "." + timeBefore.get(Calendar.MILLISECOND));
		
		UUID writeUid = UUID.randomUUID();
		
		for (int i = 0; i < nodesArray.length; i++) {
			log.debug(writeUid.toString() + ", " + i + ": Node in batch: " + nodesArray[i].toString());
		}
		
		for (int i = 0; i < valuesArray.length; i++) {
			log.debug(writeUid.toString() + ", " + i + ": Value for node: " + valuesArray[i].toString());
		}
		
		String timeBeforeText = sdf.format(Calendar.getInstance().getTime()) + "." + Calendar.getInstance().get(Calendar.MILLISECOND);


		for (int i = 0; i < nodes.size(); i++) {
			logOpcEvent(nodes.get(i), null, valuesArray[i].toString(), "write");
		}
		
		StatusCode[] status = client.writeValues(nodesArray, valuesArray);
		
		String timeAfterText = sdf.format(Calendar.getInstance().getTime()) + "." + Calendar.getInstance().get(Calendar.MILLISECOND);
		log.debug("writeValues;" + writeUid.toString() + ";" + timeBeforeText + ";" + timeAfterText + ";" + valuesArray.length + ";" + status.length);

		for (int i = 0; i < status.length; i++) {
			log.debug(i + ": " + status[i].toString());
		}

		Calendar timeAfter = Calendar.getInstance();
		log.debug("<after writeValues> time: " + sdf.format(timeAfter.getTime()) + "." + timeAfter.get(Calendar.MILLISECOND));
	}

	/**
	 * Register all nodes with transactions as subscription
	 */
	public void subscribeAll() {
		getOPCSubscriptionManager().createSubscriptions(this);
	}

	/**
	 * Returns all subscriptions registered on OPC server.
	 */
	@Override
	public List<SubscriptionItem> getSubscriptions() {
		log.debug("OPC getSubscription: Start.");

		ArrayList<SubscriptionItem> subscriptions = new ArrayList<>();
		for (ClientSubscription entry : getOPCSubscriptionManager().getSubscriptionsForClient(this, false)) {
			OPCNode opcNode = entry.getOpcNode();
			Subscription subscription = entry.getSubscriptionReference();
			log.debug("OPC getSubscription: Item found: " + Convert.toString(opcNode.getOpcNodeId()));

			String opcNodeId = opcNode.getOpcNodeId();
			String opcAttributeId = null;
			String transactionDefinitionName = opcNode.getTransactionDefinitionName();
			String subscriptionStatus = StatusCode.BAD.toString();
			DateTime subscriptionLastAlive = null;
			String publishingInterval = null;

			try {
				opcAttributeId = OPCAttributeId.findAttributeId(opcNode.getOpcNodeAttributeId()).name();
			} catch (Exception e) {
				log.warn("OPC getSubscription: Failed to read subscription properties. NodeID: " + opcNode.getOpcNodeId() + ". SubscriptionID: " + subscription.getSubscriptionId());
			}

			//subscription is connected - there is last execution, publishing interval and status
			if (Convert.isNotNull(subscription) && subscription.getItemCount() > 0) {
				if (Convert.isNotNull(subscription.getLastAlive())) {
					subscriptionLastAlive = new DateTime(subscription.getLastAlive().getTimeInMillis());
				}
				MonitoredItemBase subscriptionItem = subscription.getItems()[0];
				opcNodeId = subscriptionItem.getNodeId().toString();
				publishingInterval = Double.toString(subscription.getPublishingInterval());
				subscriptionStatus = subscription.getStatus().toString();
			}

			SubscriptionItem subscriptionItem = new SubscriptionItem(subscription.getSubscriptionId().intValue(), opcNodeId, opcAttributeId, transactionDefinitionName,
					subscriptionLastAlive, publishingInterval, subscriptionStatus);
			subscriptions.add(subscriptionItem);
		}

		log.debug("OPC getSubscription: End.");
		return subscriptions;
	}

	/**
	 * Cancel all subscriptions.
	 * @throws ServiceException
	 */
	public void unsubscribeAll() throws ServiceException {
		
		// Invalidate all internal classes for subscription (this should be managed by client, but client may be disconnected)
		getOPCSubscriptionManager().unregisterSubscriptions(this);
		
		if (this.isConnected()) {
			Subscription[] subscriptions = client.getSubscriptions();
			if (subscriptions.length > 0) {
				client.removeSubscriptions(subscriptions);
			}
		}
	}
	
	/**
	 * Registers all subscriptions to OPCSubscriptionManager
	 */
	public void registerAllSubscriptions() {
		for (OPCNode opcNode: getEvent().getListener().getOpcNodes()) {
			if (Convert.isNotBlank(opcNode.getTransactionDefinitionName())) {
				log.debug("OPCClient.registerAll ... registering NodeId: " + opcNode.getOpcNodeId());
				getOPCSubscriptionManager().registerDisconnectedSubscription(this, opcNode);
			}
		}
	}

	private enum OPCSecurityMode {

		BASIC128RSA15_SIGN(SecurityMode.BASIC128RSA15_SIGN),
		BASIC128RSA15_SIGN_ENCRYPT(SecurityMode.BASIC128RSA15_SIGN_ENCRYPT),
		BASIC256_SIGN(SecurityMode.BASIC256_SIGN),
		BASIC256_SIGN_ENCRYPT(SecurityMode.BASIC256_SIGN_ENCRYPT),
		BASIC256SHA256_SIGN_ENCRYPT(SecurityMode.BASIC256SHA256_SIGN_ENCRYPT),
		NONE(SecurityMode.NONE);

		private SecurityMode securityMode = null;

		OPCSecurityMode(SecurityMode securityMode) {
			this.securityMode = securityMode;
		}

		public SecurityMode getSecurityMode() {
			return this.securityMode;
		}

		/**
		 * Parse SecurityMode from string representation.
		 *
		 * @param securityMode
		 * @return
		 */
		public static SecurityMode parseSecurityMode(String securityMode) {
			OPCSecurityMode opcSecurityMode = securityMode != null ? OPCSecurityMode.valueOf(securityMode) : OPCSecurityMode.NONE;
			return opcSecurityMode.getSecurityMode();
		}

	}

	/**
	 * Returns a list of subordinate nodes in form: <nodeId, nodeType>
	 * @param parentNodeIdString
	 * @return
	 */
	public HashList<String, String> readChildNodes(String parentNodeIdString) {
		SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT_MASK);
		NodeId nodeId = NodeId.parseNodeId(parentNodeIdString);
		
		List<ReferenceDescription> childNodes = null;
		
		Calendar timeBefore = Calendar.getInstance();
		log.debug("<before readChild> nodeId: " + parentNodeIdString + TIME + sdf.format(timeBefore.getTime()) + "." 
				+ timeBefore.get(Calendar.MILLISECOND));
		try {
			childNodes = client.getAddressSpace().browse(nodeId);
		} catch (ServiceException | StatusException e) {
			log.error("Cannot read subordinate nodes for nodeId: " + parentNodeIdString, e);
		}

		HashList<String, String> subordinateNodes = new HashList<String, String>();
		
		for (ReferenceDescription childNode: childNodes) {
			String childNodeId = "";
			if (childNode.getNodeId().getNamespaceIndex() > 0) {
				childNodeId += "ns=" + childNode.getNodeId().getNamespaceIndex() + ";s=";
			} else {
				childNodeId += "i=";
			}
			childNodeId += childNode.getNodeId().getValue();
			if (!subordinateNodes.containsKey(childNodeId)) {
			subordinateNodes.put(childNodeId, childNode.getNodeClass().name());
		}
		}
		
		Calendar timeAfter = Calendar.getInstance();
		log.debug("<after readChild> nodeId: " + parentNodeIdString + TIME + sdf.format(timeAfter.getTime()) + "." 
				+ timeAfter.get(Calendar.MILLISECOND));
		return subordinateNodes;
	}

	public String readNodeDataType(String pNodeId) throws ServiceException, StatusException {
		return client.readAttribute(NodeId.parseNodeId(pNodeId), Attributes.DataType).getValue().getValue().toString();
	}

	/**
	 *
	 * @param code
	 * @return
	 */
	public OPCDataType getDataTypeByIdCode(String code) {
		
		for (OPCDataType dataType: OPCDataTypeStandard.values()) {
			if (dataType.getIdCode().equals(code)) {
				return dataType;
			}
		}
		
		return OPCDataType.NotImplemented;
	}
	
	/**
	 * Log OPC event to database according settings on OPC node additional attributes.
	 * TCPIPlistener service has to run in the same company as ResponseDurationMachineLogger service.
	 * 
	 * @param opcNode
	 * @param machineCode
	 * @param value
	 * @param author
	 */
	public void logOpcEvent(String opcNode, String machineCode, String value, String author) {
		TCPIPListenerEvent event = getEvent();
		MultiEventService service = event.getService();

		event.getCoreGate().saveMachineResponseDuration(service.getCode(), service.getCompanyId(), event.getCode(), opcNode, machineCode, value, author);
	}

	@Override
	public Logger getLogger() {
		return log;
	}
}
