package cz.aimtec.dci.coreservices.listener.tcpip.opc;

import com.prosysopc.ua.ServiceException;
import com.prosysopc.ua.StatusException;
import com.prosysopc.ua.client.Subscription;
import cz.aimtec.dci.coreservices.api.model.ServiceConfiguration;
import cz.aimtec.dci.coreservices.api.model.ServiceEventConfiguration;
import cz.aimtec.dci.coreservices.implementation.TCPIPListenerEvent;
import cz.aimtec.dci.coreservices.listener.ListenerType;
import cz.aimtec.dci.coreservices.listener.MessageListener;
import cz.aimtec.dci.coreservices.listener.tcpip.TCPIPClient;
import cz.aimtec.dci.coreservices.service.ServiceEvent;
import cz.aimtec.dci.coreservices.service.ServicesDispatcher;
import cz.aimtec.util.Convert;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

/**
 * Global storage and manager for OPC subscriptions
 * @author asel
 *
 */
@Component
public class OPCSubscriptionManager {

	@Autowired
	private ServicesDispatcher	servicesDispatcher;

	private static Logger log = Logger.getLogger(OPCSubscriptionManager.class);
	private Map<String, List<ClientSubscription>> clientSubscriptions = new ConcurrentHashMap<>();


	/**
	 *
	 * @param client
	 */
	public void createSubscriptions(OPCClient client) {
		List<ClientSubscription> subscriptions = findSubscriptionsForClient(client.getEventCode());
		for (ClientSubscription clientSubscription : subscriptions) {
			OPCNode opcNode = clientSubscription.getOpcNode();
			if (doesSubscriptionAlreadyExist(client, opcNode)) {
				// node's been already registered
				TCPIPClient.registerListenerCodeForLogging(client);
				log.error("Double subscription detected! NodeId: " + opcNode.getOpcNodeId() + ", client: " + client.getEvent().getCode());
				return;
			}

			if (Convert.isNotBlank(opcNode.getTransactionDefinitionName())) {
				try {
					client.subscribe(opcNode);
				} catch (ServiceException | StatusException e) {
				log.error("Error during creating subscriptions.", e);
				}
			}	
		}
	}
	
	/**
	 * Register disconnected(unsubscribed) item to the manager without detailed info
	 * @param client - reference to OPCClient
	 * @param opcNode - opcNode event 
	 */
	public void registerDisconnectedSubscription(OPCClient client, OPCNode opcNode) {
		if (doesSubscriptionAlreadyExist(client, opcNode)) {
			log.error("Double registration detected! NodeId: " + opcNode.getOpcNodeId() + ", client: " + client.getEvent().getCode());
			return;
		}

		if (!clientSubscriptions.containsKey(client.getEventCode())) {
			clientSubscriptions.put(client.getEventCode(), new ArrayList<>());
		}

		//Just register future subscription
		if (Convert.isNotBlank(opcNode.getTransactionDefinitionName())) {
			this.clientSubscriptions.get(client.getEventCode()).add(new ClientSubscription(client, null, null, opcNode));
		}
	}


	/**
	 * @param opcNode
	 * @param subscriptionReference
	 * @param subscriptionListener
	 */
	public void setSubscriptionInfo(OPCClient client, OPCNode opcNode, Subscription subscriptionReference, DefaultTransMonitoredDataItemListener subscriptionListener) {
		List<ClientSubscription> subscriptions = findSubscriptionsForClient(client.getEventCode());

		// find subscription
		Optional<ClientSubscription> clientSubscriptionOptional = subscriptions.stream()
				.filter(subscription -> subscription.getOpcNode().equals(opcNode)).findFirst();

		if (clientSubscriptionOptional.isPresent()) {
			ClientSubscription clientSubscription = clientSubscriptionOptional.get();
			clientSubscription.setSubscriptionReference(subscriptionReference);
			clientSubscription.setSubscriptionThread(subscriptionListener);
		} else {
			log.error("There is subscription for node! NodeID:" + opcNode.getOpcNodeId() + "didn't register to OPCSubscriptionManager");
		}
	}
	
	/**
	 * Stop all subscriptions from given OPC client.
	 * @param opcClient
	 */
	public void stopSubscriptions(OPCClient opcClient) {
		try {
			opcClient.unsubscribeAll();
		} catch (ServiceException e) {
			TCPIPClient.registerListenerCodeForLogging(opcClient);
			log.error("Error during stopping subscriptions.", e);
		}
	}
	
	/**
	 * Invalidate all threads that belong to given OPC client
	 * @param opcClient
	 */
	public void unregisterSubscriptions(OPCClient opcClient) {
		List<ClientSubscription> subscriptions = findSubscriptionsForClient(opcClient.getEventCode());
		for (ClientSubscription subscription : subscriptions) {
			if (subscription.getSubscriptionThread() != null) {
				subscription.getSubscriptionThread().invalidate();
			}
		}

		removeAllInvalidatedSubscriptions(subscriptions);
	}

	/**
	 * Return subscriptions for specific OPC Client (that don't have to be connected)
	 *
	 * @param opcClient
	 * @param validOnly
	 * @return
	 */
	public List<ClientSubscription> getSubscriptionsForClient(OPCClient opcClient, boolean validOnly) {
		List<ClientSubscription> subscriptions = findSubscriptionsForClient(opcClient.getEventCode());
		return subscriptions.stream()
				.filter(subscription -> subscription.getClient().equals(opcClient)
						&& (!validOnly || (subscription.getSubscriptionThread() != null && subscription.getSubscriptionThread().isValid())))
				.collect(Collectors.toList());
	}
	
	/**
	 * Unsubscribe all subscriptions from all opc servers
	 */
	private void unsubscribeAllSubscriptions(String pServiceCode, Integer pCompanyId) {
		// Cancel all subscriptions
		for (ServiceEvent serviceEvent : servicesDispatcher.getServiceEvents(pServiceCode, pCompanyId)) {
			TCPIPListenerEvent event = (TCPIPListenerEvent) serviceEvent;
			
			if (event.getListener().getListenerType().equals(ListenerType.OPC)) {
				try {
					((OPCClient) event.getClient()).unsubscribeAll();
				} catch (ServiceException e) {
					TCPIPClient.registerListenerCodeForLogging(event.getClient());
					log.error("Error during unsubscribe subscriptions.", e);
				}
			}
		}
	}
	
	/**
	 * Make subscriptions for all opn nodes in cache.
	 */
	private void registerAllSubscriptions(ServiceConfiguration pServiceConfiguration) {
		String serviceCode = pServiceConfiguration.getServiceCode();
		Integer companyId = pServiceConfiguration.getCompanyId();
		MessageListener listener;
		OPCClient client;
		ServiceEventConfiguration eventConfiguration;
		MessageListener listenerTemplate;
		
		for (ServiceEvent serviceEvent : servicesDispatcher.getServiceEvents(serviceCode, companyId)) {
			TCPIPListenerEvent event = (TCPIPListenerEvent) serviceEvent;
			listener = event.getListener();
			
			// ------------------- event configuration
			eventConfiguration = null;
			for (ServiceEventConfiguration eventConfig : pServiceConfiguration.getEvents()) {
				if (eventConfig.getCode().equals(event.getCode())) {
					eventConfiguration = eventConfig;
					break;
				}
			}
			
			if (eventConfiguration != null) {
				// ----------------- nodes reinitialization
				listenerTemplate = new MessageListener(eventConfiguration.getListener());
				listener.getOpcNodes().clear();
				listener.getOpcNodes().addAll(listenerTemplate.getOpcNodes());

				
				// -------------------- nodes subscription 
				if (listener.getListenerType().isOPCType()) {
					client = (OPCClient) event.getClient();
					client.registerAllSubscriptions();
					client.subscribeAll();
				}
			}
		}
	}

	/**
	 * Reload all subscriptions.
	 *
	 * @param pServiceConfiguration new service configuration
	 */
	public void refreshSubscriptions(ServiceConfiguration pServiceConfiguration) {
		this.unsubscribeAllSubscriptions(pServiceConfiguration.getServiceCode(), pServiceConfiguration.getCompanyId());
		this.registerAllSubscriptions(pServiceConfiguration);
	}

	private void removeAllInvalidatedSubscriptions(List<ClientSubscription> subscriptions) {
		for (Iterator<ClientSubscription> iterator = subscriptions.iterator(); iterator.hasNext(); ) {
			ClientSubscription subscription = iterator.next();
			if (subscription.getSubscriptionThread() == null || !subscription.getSubscriptionThread().isValid()) {
				iterator.remove();
			}
		}
	}

	private boolean doesSubscriptionAlreadyExist(OPCClient opcClient, OPCNode opcNode) {
		List<ClientSubscription> subscriptions = findSubscriptionsForClient(opcClient.getEventCode());
		for (ClientSubscription clientSubscription : subscriptions) {
			if (clientSubscription.getOpcNode().equals(opcNode)) {
				Subscription s = clientSubscription.getSubscriptionReference();
				if (Convert.isNotNull(s) && s.isAlive()) {
					return true;
				}
			}
		}

		return false;
	}

	private List<ClientSubscription> findSubscriptionsForClient(String opcClientCode) {
		if (!this.clientSubscriptions.containsKey(opcClientCode)) {
			this.clientSubscriptions.put(opcClientCode, new ArrayList<>());
		}

		return clientSubscriptions.get(opcClientCode);
	}
}
