package cz.aimtec.dci.microservice.client;

import java.util.Map;

import javax.ws.rs.core.MediaType;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import cz.aimtec.dci.microservice.discovery.DiscoveryService;
import cz.aimtec.dci.microservice.discovery.ServiceLocationProperties;
import cz.aimtec.dci.microservice.eventbus.ClientResponseHandler;
import cz.aimtec.dci.microservice.eventbus.EventBusClient;
import cz.aimtec.dci.microservice.eventbus.RequestMessage;

/**
 * Service client supporting method asynchronous method call.
 *
 * @author	hors
 * @version	16. 9. 2020/hors - created<br>
 */
public class ServiceClientAsync extends ServiceClient {
	
	private static Logger			logger = LoggerFactory.getLogger(ServiceClientAsync.class);

	protected final EventBusClient	eventBusClient;


	/**
	 * Basic constructor.
	 *
	 * @param serviceName			destination service name
	 * @param version				destination service version
	 * @param discoveryService		discovery service
	 */
	public ServiceClientAsync(DiscoveryService discoveryService, EventBusClient eventBusClient, String serviceName, String version) {
		super(discoveryService, serviceName, version);
		this.eventBusClient = eventBusClient;
	}

	public ServiceClientAsync(DiscoveryService discoveryService, EventBusClient eventBusClient, String serviceName) {
		super(discoveryService, serviceName);
		this.eventBusClient = eventBusClient;
	}

	public ServiceClientAsync(DiscoveryService discoveryService) {
		super(discoveryService);
		this.eventBusClient = null;
	}

	/**
	 * Basic constructor produces incomplete instance.
	 */
	public ServiceClientAsync() {
		super(null);
		this.eventBusClient = null;
	}
	

	/**
	 * Service method asynchronous call based on service discovery.
	 * It creates message using parameters and sends it to related endpoint request queue. 
	 *
	 * @param pMethod       	HTTP method
	 * @param pPath            	service context path
	 * @param pQueryParameters  additional query parameters
	 * @param pRequestBody		message body object instance
	 * @param pCookies			request cookies collection
	 * @param pRequestMediaType body media type string
	 * @param pAcceptMediaType	{@code Accept} request HTTP header attribute
	 * @param pResponseHandler	response handler 
	 */
	public void callMethodAsync(HttpMethod pMethod,
	                            String pPath,
	                            Map<String, Object> pQueryParameters,
	                            String pRequestBody,
	                            Map<String, String> pCookies,
	                            MediaType pRequestMediaType,
	                            MediaType pAcceptMediaType,
	                            ClientResponseHandler pResponseHandler) {
		// ------------------------ service discovery
		ServiceLocationProperties resolvedServiceInstance = findAppropriateServiceInstance();

		// ----------------------- calling the method
		if (resolvedServiceInstance != null) {
			String path = resolvePath(resolvedServiceInstance, pPath);
			
			logger.debug("Service asynchronous call: {} - required version: {}", pPath, getRequiredVersion());
			sendRequestAsync(pMethod, path, pQueryParameters, pRequestBody, pCookies, pRequestMediaType, pAcceptMediaType, pResponseHandler);
		}
	}
	
	/**
	 * Service method asynchronous call based on service discovery.
	 * It creates message using parameters and sends it to related endpoint request queue. 
	 *
	 * @param pMethod       	HTTP method
	 * @param pPath            	service context path
	 * @param pQueryParameters  additional query parameters
	 * @param pRequestBody		message body object instance
	 * @param pCookies			request cookies collection
	 * @param pResponseHandler	response handler 
	 */
	public void callMethodAsync(HttpMethod pMethod,
	                            String pPath,
	                            Map<String, Object> pQueryParameters,
	                            Object pRequestBody,
	                            Map<String, String> pCookies,
	                            ClientResponseHandler pResponseHandler) {
		callMethodAsync(pMethod,
		                pPath,
		                pQueryParameters,
		                ServiceClient.toJsonString(pRequestBody),
		                pCookies,
		                ServiceClient.DEFAULT_CONTENT_TYPE,
		                ServiceClient.DEFAULT_CONTENT_TYPE,
		                pResponseHandler);
	}
	
	/**
	 * General asynchronous request sending.
	 *
	 * @param pMethod       	HTTP method
	 * @param pPath            	service context path
	 * @param pQueryParameters  additional query parameters
	 * @param pRequestBody		message body object instance
	 * @param pCookies			request cookies collection
	 * @param pRequestMediaType body media type string
	 * @param pAcceptMediaType	{@code Accept} request HTTP header attribute
	 * @param pResponseHandler	response handler 
	 */
	public void sendRequestAsync(HttpMethod pMethod,
	                             String pPath,
	                             Map<String, Object> pQueryParameters,
	                             String pRequestBody,
	                             Map<String, String> pCookies,
								 MediaType pRequestMediaType,
								 MediaType pAcceptMediaType,
	                             ClientResponseHandler pResponseHandler) {
		// --------------------------- sending the request
		logger.debug("Sending asynchronous request: {}, {}, {}, {}, {}", pPath, pMethod, pQueryParameters, pRequestBody, pCookies);
		String requestId = eventBusClient.sendClientRequestMessage(pPath,
		                                                           new RequestMessage(pMethod,
		                                                                              pQueryParameters,
		                                                                              pRequestBody,
		                                                                              pCookies,
		                                                                              pRequestMediaType.toString(),
		                                                                              pAcceptMediaType.toString()));

		// --------------------------- subscribing for response
		eventBusClient.registerClientResponseMessageHandler(pPath, requestId, pResponseHandler);
	}
}
